# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Interactive API package."""

from logging import getLogger

import numpy as np

from openfl.utilities import change_tags
from openfl.utilities import split_tensor_dict_for_holdouts
from openfl.utilities import TensorKey


class CoreTaskRunner:
    """Federated Learning Task Runner Class."""

    def _prepare_tensorkeys_for_agggregation(self, metric_dict, validation_flag,
                                             col_name, round_num):
        """
        Prepare tensorkeys for aggregation.

        Returns (global_tensor_dict, local_tensor_dict)
        """
        global_tensor_dict, local_tensor_dict = {}, {}
        origin = col_name
        if not validation_flag:
            # Output metric tensors (scalar)
            tags = ('trained',)

            # output model tensors (Doesn't include TensorKey)
            output_model_dict = self.get_tensor_dict(with_opt_vars=True)
            global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
                self.logger, output_model_dict,
                **self.tensor_dict_split_fn_kwargs
            )

            # Create global tensorkeys
            global_tensorkey_model_dict = {
                TensorKey(tensor_name, origin, round_num, False, tags):
                    nparray for tensor_name, nparray in global_model_dict.items()}
            # Create tensorkeys that should stay local
            local_tensorkey_model_dict = {
                TensorKey(tensor_name, origin, round_num, False, tags):
                    nparray for tensor_name, nparray in local_model_dict.items()}
            # The train/validate aggregated function of the next
            # round will look for the updated model parameters.
            # This ensures they will be resolved locally
            next_local_tensorkey_model_dict = {TensorKey(
                tensor_name, origin, round_num + 1, False, ('model',)): nparray
                for tensor_name, nparray in local_model_dict.items()}

            global_tensor_dict = global_tensorkey_model_dict
            local_tensor_dict = {**local_tensorkey_model_dict, **next_local_tensorkey_model_dict}

            # Update the required tensors if they need to be
            # pulled from the aggregator
            # TODO this logic can break if different collaborators
            #  have different roles between rounds.
            # For example, if a collaborator only performs validation
            # in the first round but training
            # in the second, it has no way of knowing the optimizer
            # state tensor names to request from the aggregator
            # because these are only created after training occurs.
            # A work around could involve doing a single epoch of training
            # on random data to get the optimizer names,
            # and then throwing away the model.
            if self.opt_treatment == 'CONTINUE_GLOBAL':
                self.initialize_tensorkeys_for_functions(with_opt_vars=True)

            # This will signal that the optimizer values are now present,
            # and can be loaded when the model is rebuilt
            self.training_round_completed = True

        else:
            suffix = 'validate' + validation_flag
            tags = (suffix,)
        tags = change_tags(tags, add_field='metric')
        metric_dict = {
            TensorKey(metric, origin, round_num, True, tags):
                np.array(value) for metric, value in metric_dict.items()
        }
        global_tensor_dict = {**global_tensor_dict, **metric_dict}

        return global_tensor_dict, local_tensor_dict

    def adapt_tasks(self):
        """
        Prepare tasks for the collaborator.

        Using functions from a task provider (deserialized interface object) and
        registered task contracts prepares callable tasks to be invoked by the collaborator.

        Preparing includes conditional model rebuilding and filling output dicts
        with tensors for aggregation and storing in local DB.

        There is an assumption that any training task accepts optimizer as one
        of the arguments, thus the model should be aggregated after such tasks.
        """

        def task_binder(task_name, callable_task):
            def collaborator_adapted_task(col_name, round_num, input_tensor_dict, **kwargs):
                task_contract = self.task_provider.task_contract[task_name]
                # Validation flag can be [False, '_local', '_agg']
                validation_flag = True if task_contract['optimizer'] is None else False
                task_settings = self.task_provider.task_settings[task_name]

                device = kwargs.get('device', 'cpu')

                self.rebuild_model(input_tensor_dict, validation=validation_flag, device=device)
                task_kwargs = {}
                if validation_flag:
                    loader = self.data_loader.get_valid_loader()
                    if kwargs['apply'] == 'local':
                        validation_flag = '_local'
                    else:
                        validation_flag = '_agg'
                else:
                    loader = self.data_loader.get_train_loader()
                    # If train task we also pass optimizer
                    task_kwargs[task_contract['optimizer']] = self.optimizer

                if task_contract['round_num'] is not None:
                    task_kwargs[task_contract['round_num']] = round_num

                for en_name, entity in zip(['model', 'data_loader', 'device'],
                                           [self.model, loader, device]):
                    task_kwargs[task_contract[en_name]] = entity

                # Add task settings to the keyword arguments
                task_kwargs.update(task_settings)

                # Here is the training metod call
                metric_dict = callable_task(**task_kwargs)

                return self._prepare_tensorkeys_for_agggregation(
                    metric_dict, validation_flag, col_name, round_num)

            return collaborator_adapted_task

        for task_name, callable_task in self.task_provider.task_registry.items():
            self.TASK_REGISTRY[task_name] = task_binder(task_name, callable_task)

    def __init__(self, **kwargs):
        """
        Initialize Task Runner.

        This class is a part of the Interactive python API release.
        It is no longer a user interface entity that should be subclassed
        but a part of OpenFL kernel.
        """
        self.set_logger()

        self.kwargs = kwargs

        self.TASK_REGISTRY = {}

        # Why is it here
        self.opt_treatment = 'RESET'
        self.tensor_dict_split_fn_kwargs = {}
        self.required_tensorkeys_for_function = {}

        # Complete hell below
        self.training_round_completed = False
        # overwrite attribute to account for one optimizer param (in every
        # child model that does not overwrite get and set tensordict) that is
        # not a numpy array
        self.tensor_dict_split_fn_kwargs.update({
            'holdout_tensor_names': ['__opt_state_needed']
        })

    def set_task_provider(self, task_provider):
        """
        Set task registry.

        This method recieves Task Interface object as an argument
        and uses provided callables and information to prepare
        tasks that may be called by the collaborator component.
        """
        if task_provider is None:
            return
        self.task_provider = task_provider
        self.adapt_tasks()

    def set_data_loader(self, data_loader):
        """Register a data loader initialized with local data path."""
        self.data_loader = data_loader

    def set_model_provider(self, model_provider):
        """Retrieve a model and an optimizer from the interface object."""
        self.model_provider = model_provider
        self.model = self.model_provider.provide_model()
        self.optimizer = self.model_provider.provide_optimizer()

    def set_framework_adapter(self, framework_adapter):
        """
        Set framework adapter.

        Setting a framework adapter allows first extraction of the weigths
        of the model with the purpose to make a list of parameters to be aggregated.
        """
        self.framework_adapter = framework_adapter
        if self.opt_treatment == 'CONTINUE_GLOBAL':
            aggregate_optimizer_parameters = True
        else:
            aggregate_optimizer_parameters = False
        self.initialize_tensorkeys_for_functions(with_opt_vars=aggregate_optimizer_parameters)

    def set_logger(self):
        """Set up the log object."""
        self.logger = getLogger(__name__)

    def set_optimizer_treatment(self, opt_treatment):
        # SHould be removed! We have this info at the initialization time
        # and do not change this one during training.
        """Change the treatment of current instance optimizer."""
        self.opt_treatment = opt_treatment

    def rebuild_model(self, input_tensor_dict, validation=False, device='cpu'):
        """
        Parse tensor names and update weights of model. Handles the optimizer treatment.

        Returns:
            None
        """
        if self.opt_treatment == 'RESET':
            self.reset_opt_vars()
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False, device=device)
        elif (self.training_round_completed
              and self.opt_treatment == 'CONTINUE_GLOBAL' and not validation):
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=True, device=device)
        else:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False, device=device)

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """
        Get the required tensors for specified function that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.

        Parameters
        ----------
        None

        Returns
        -------
        List
            [TensorKey]
        """
        # We rely on validation type tasks parameter `apply`
        # In the interface layer we add those parameters automatically
        if 'apply' not in kwargs:
            return [
                TensorKey(tensor_name, 'GLOBAL', 0, False, ('model',))
                for tensor_name in self.required_tensorkeys_for_function['global_model_dict']
            ] + [
                TensorKey(tensor_name, 'LOCAL', 0, False, ('model',))
                for tensor_name in self.required_tensorkeys_for_function['local_model_dict']
            ]

        if kwargs['apply'] == 'local':
            return [
                TensorKey(tensor_name, 'LOCAL', 0, False, ('trained',))
                for tensor_name in {
                    **self.required_tensorkeys_for_function['local_model_dict_val'],
                    **self.required_tensorkeys_for_function['global_model_dict_val']
                }
            ]

        elif kwargs['apply'] == 'global':
            return [
                TensorKey(tensor_name, 'GLOBAL', 0, False, ('model',))
                for tensor_name in self.required_tensorkeys_for_function['global_model_dict_val']
            ] + [
                TensorKey(tensor_name, 'LOCAL', 0, False, ('model',))
                for tensor_name in self.required_tensorkeys_for_function['local_model_dict_val']
            ]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """Set the required tensors for all publicly accessible task methods.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function.

        Args:
            None

        Returns:
            None
        """
        # TODO: Framework adapters should have separate methods for dealing with optimizer
        # Set model dict for validation tasks
        output_model_dict = self.get_tensor_dict(with_opt_vars=False)
        global_model_dict_val, local_model_dict_val = split_tensor_dict_for_holdouts(
            self.logger,
            output_model_dict,
            **self.tensor_dict_split_fn_kwargs
        )
        # Now set model dict for training tasks
        if with_opt_vars:
            output_model_dict = self.get_tensor_dict(with_opt_vars=True)
            global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
                self.logger, output_model_dict,
                **self.tensor_dict_split_fn_kwargs
            )
        else:
            global_model_dict = global_model_dict_val
            local_model_dict = local_model_dict_val

        self.required_tensorkeys_for_function['global_model_dict'] = global_model_dict
        self.required_tensorkeys_for_function['local_model_dict'] = local_model_dict
        self.required_tensorkeys_for_function['global_model_dict_val'] = global_model_dict_val
        self.required_tensorkeys_for_function['local_model_dict_val'] = local_model_dict_val

    def reset_opt_vars(self):
        """
        Reset optimizer variables.

        Resets the optimizer variables

        """
        self.optimizer = self.model_provider.provide_optimizer()

    def get_train_data_size(self):
        """
        Get the number of training examples.

        It will be used for weighted averaging in aggregation.

        Returns:
            int: The number of training examples.
        """
        return self.data_loader.get_train_data_size()

    def get_valid_data_size(self):
        """
        Get the number of examples.

        It will be used for weighted averaging in aggregation.

        Returns:
            int: The number of validation examples.
        """
        return self.data_loader.get_valid_data_size()

    def get_tensor_dict(self, with_opt_vars=False):
        """Return the tensor dictionary.

        Args:
            with_opt_vars (bool): Return the tensor dictionary including the
                                  optimizer tensors (Default=False)

        Returns:
            dict: Tensor dictionary {**dict, **optimizer_dict}

        """
        args = [self.model]
        if with_opt_vars:
            args.append(self.optimizer)

        return self.framework_adapter.get_tensor_dict(*args)

    def set_tensor_dict(self, tensor_dict, with_opt_vars=False, device='cpu'):
        """Set the tensor dictionary.

        Args:
            tensor_dict: The tensor dictionary
            with_opt_vars (bool): Return the tensor dictionary including the
                                  optimizer tensors (Default=False)

        """
        # Sets tensors for model layers and optimizer state.
        # FIXME: self.parameters() instead? Unclear if load_state_dict() or
        #  simple assignment is better
        # for now, state dict gives us names, which is good
        # FIXME: do both and sanity check each time?
        args = [self.model, tensor_dict]
        if with_opt_vars:
            args.append(self.optimizer)

        kwargs = {'device': device, }

        return self.framework_adapter.set_tensor_dict(*args, **kwargs)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregation functions core package."""

from .adaptive_aggregation import AdaptiveAggregation
from .interface import AggregationFunction

__all__ = ['AggregationFunction',
           'AdaptiveAggregation']



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""KCPipeline module."""


import copy as co
import gzip as gz

import numpy as np
from sklearn import cluster

from .pipeline import TransformationPipeline
from .pipeline import Transformer


class KmeansTransformer(Transformer):
    """A K-means transformer class to quantize input data."""

    def __init__(self, n_cluster=6):
        """Class initializer.

        Args:
            n_cluster (int): Number of clusters for the K-means
        """
        self.lossy = True
        self.n_cluster = n_cluster

    def forward(self, data, **kwargs):
        """
        Quantize data into n_cluster levels of values.

        Args:
            data: an numpy array from the model tensor_dict
            data: an numpy array being quantized
            **kwargs: Variable arguments to pass
        """
        metadata = {'int_list': list(data.shape)}
        # clustering
        k_means = cluster.KMeans(n_clusters=self.n_cluster, n_init=self.n_cluster)
        data = data.reshape((-1, 1))
        if data.shape[0] >= self.n_cluster:
            k_means = cluster.KMeans(
                n_clusters=self.n_cluster, n_init=self.n_cluster)
            k_means.fit(data)
            quantized_values = k_means.cluster_centers_.squeeze()
            indices = k_means.labels_
            quant_array = np.choose(indices, quantized_values)
        else:
            quant_array = data

        int_array, int2float_map = self._float_to_int(quant_array)
        metadata['int_to_float'] = int2float_map

        return int_array, metadata

    def backward(self, data, metadata, **kwargs):
        """Recover data array back to the original numerical type and the shape.

        Args:
            data: an flattened numpy array
            metadata: dictionary to contain information for recovering ack to original data array

        Returns:
            data: Numpy array with original numerical type and shape
        """
        # convert back to float
        # TODO
        data = co.deepcopy(data)
        int2float_map = metadata['int_to_float']
        for key in int2float_map:
            indices = data == key
            data[indices] = int2float_map[key]
        data_shape = list(metadata['int_list'])
        data = data.reshape(data_shape)
        return data

    @staticmethod
    def _float_to_int(np_array):
        """Create look-up table for conversion between floating and integer types.

        Args:
            np_array: A Numpy array

        Returns:
            int_array: The input Numpy float array converted to an integer array
            int_to_float_map
        """
        flatten_array = np_array.reshape(-1)
        unique_value_array = np.unique(flatten_array)
        int_array = np.zeros(flatten_array.shape, dtype=np.int)
        int_to_float_map = {}
        float_to_int_map = {}
        # create table
        for idx, u_value in enumerate(unique_value_array):
            int_to_float_map.update({idx: u_value})
            float_to_int_map.update({u_value: idx})
            # assign to the integer array
            indices = np.where(flatten_array == u_value)
            int_array[indices] = idx
        int_array = int_array.reshape(np_array.shape)
        return int_array, int_to_float_map


class GZIPTransformer(Transformer):
    """A GZIP transformer class to losslessly compress data."""

    def __init__(self):
        """Initialize."""
        self.lossy = False

    def forward(self, data, **kwargs):
        """Compress data into bytes.

        Args:
            data: A Numpy array

        Returns:
            GZIP compressed data object
        """
        bytes_ = data.astype(np.float32).tobytes()
        compressed_bytes_ = gz.compress(bytes_)
        metadata = {}
        return compressed_bytes_, metadata

    def backward(self, data, metadata, **kwargs):
        """Decompress data into numpy of float32.

        Args:
            data: Compressed GZIP data
            metadata:
            **kwargs: Additional parameters to pass to the function

        Returns:
            data: Numpy array
        """
        decompressed_bytes_ = gz.decompress(data)
        data = np.frombuffer(decompressed_bytes_, dtype=np.float32)
        return data


class KCPipeline(TransformationPipeline):
    """A pipeline class to compress data lossly using k-means and GZIP methods."""

    def __init__(self, p_sparsity=0.01, n_clusters=6, **kwargs):
        """Initialize a pipeline of transformers.

        Args:
            p_sparsity (float): Amount of sparsity for compression (Default = 0.01)
            n_clusters (int): Number of K-mean cluster

        Return:
            Transformer class object
        """
        # instantiate each transformer
        self.p = p_sparsity
        self.n_cluster = n_clusters
        transformers = [KmeansTransformer(self.n_cluster), GZIPTransformer()]
        super(KCPipeline, self).__init__(transformers=transformers, **kwargs)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl.federated package."""

import pkgutil
from .plan import Plan  # NOQA
from .task import TaskRunner  # NOQA
from .data import DataLoader  # NOQA

if pkgutil.find_loader('tensorflow'):
    from .task import TensorFlowTaskRunner, KerasTaskRunner, FederatedModel  # NOQA
    from .data import TensorFlowDataLoader, KerasDataLoader, FederatedDataSet  # NOQA
if pkgutil.find_loader('torch'):
    from .task import PyTorchTaskRunner, FederatedModel  # NOQA
    from .data import PyTorchDataLoader, FederatedDataSet  # NOQA
if pkgutil.find_loader('torch') and pkgutil.find_loader('tensorflow'):
    from .task import FastEstimatorTaskRunner  # NOQA
    from .data import FastEstimatorDataLoader  # NOQA



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Adagrad adaptive aggregation module."""

from typing import Dict
from typing import Optional

import numpy as np

from openfl.utilities.optimizers.numpy import NumPyAdagrad
from .core import AdaptiveAggregation
from .core import AggregationFunction
from .weighted_average import WeightedAverage


DEFAULT_AGG_FUNC = WeightedAverage()


class AdagradAdaptiveAggregation(AdaptiveAggregation):
    """Adagrad adaptive Federated Aggregation funtcion."""

    def __init__(
        self,
        *,
        agg_func: AggregationFunction = DEFAULT_AGG_FUNC,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        initial_accumulator_value: float = 0.1,
        epsilon: float = 1e-10,
    ) -> None:
        """Initialize.

        Args:
            agg_func: Aggregate function for aggregating
                parameters that are not inside the optimizer (default: WeightedAverage()).
            params: Parameters to be stored for optimization.
            model_interface: Model interface instance to provide parameters.
            learning_rate: Tuning parameter that determines
                the step size at each iteration.
            initial_accumulator_value: Initial value for squared gradients.
            epsilon: Value for computational stability.
        """
        opt = NumPyAdagrad(params=params,
                           model_interface=model_interface,
                           learning_rate=learning_rate,
                           initial_accumulator_value=initial_accumulator_value,
                           epsilon=epsilon)
        super().__init__(opt, agg_func)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Assigner module."""


class Assigner:
    r"""
    The task assigner maintains a list of tasks.

    Also it decides the policy for which collaborator should run those tasks.
    There may be many types of policies implemented, but a natural place to start is with a:

    RandomGroupedTaskAssigner  - Given a set of task groups, and a percentage,
                                 assign that task group to that
                                 percentage of collaborators in the federation.
                                 After assigning the tasks to
                                 collaborator, those tasks should be carried
                                 out each round (no reassignment
                                 between rounds)
    GroupedTaskAssigner - Given task groups and a list of collaborators that
                          belong to that task group,
                          carry out tasks for each round of experiment

    Args:
        tasks* (list of object): list of tasks to assign.
        authorized_cols (list of str): collaborators.
        rounds_to_train (int): number of training rounds.

    Note:
        \* - ``tasks`` argument is taken from ``tasks`` section of FL plan YAML file.
    """

    def __init__(self, tasks, authorized_cols,
                 rounds_to_train, **kwargs):
        """Initialize."""
        self.tasks = tasks
        self.authorized_cols = authorized_cols
        self.rounds = rounds_to_train
        self.all_tasks_in_groups = []

        self.task_group_collaborators = {}
        self.collaborators_for_task = {}
        self.collaborator_tasks = {}

        self.define_task_assignments()

    def define_task_assignments(self):
        """Abstract method."""
        raise NotImplementedError

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Abstract method."""
        raise NotImplementedError

    def get_collaborators_for_task(self, task_name, round_number):
        """Abstract method."""
        raise NotImplementedError

    def get_all_tasks_for_round(self, round_number):
        """
        Return tasks for the current round.

        Currently all tasks are performed on each round,
        But there may be a reason to change this.
        """
        return self.all_tasks_in_groups

    def get_aggregation_type_for_task(self, task_name):
        """Extract aggregation type from self.tasks."""
        if 'aggregation_type' not in self.tasks[task_name]:
            return None
        return self.tasks[task_name]['aggregation_type']



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl path checks."""

import os
from pathlib import Path
from typing import Union


def is_directory_traversal(directory: Union[str, Path]) -> bool:
    """Check for directory traversal."""
    cwd = os.path.abspath(os.getcwd())
    requested_path = os.path.relpath(directory, start=cwd)
    requested_path = os.path.abspath(requested_path)
    common_prefix = os.path.commonprefix([requested_path, cwd])
    return common_prefix != cwd



from importlib import import_module
from os.path import splitext

from .random_grouped_assigner import RandomGroupedAssigner


class AdaBoostAssigner(RandomGroupedAssigner):

    def __init__(self, task_groups, **kwargs):
        super().__init__(task_groups, **kwargs)

    def get_aggregation_type_for_task(self, task_name):
        """Extract aggregation type from self.tasks."""
        if 'aggregation_type' not in self.tasks[task_name]:
            return None

        template = self.tasks[task_name]['aggregation_type']
        # TODO: this should be integrated better with aggregator (and the eventual parameter should be handled better)
        if isinstance(template, str):
            class_name = splitext(template)[1].strip('.')
            module_path = splitext(template)[0]
            module = import_module(module_path)
            if 'n_classes' in self.tasks[task_name]:
                template = getattr(module, class_name)(self.tasks[task_name]['n_classes'])
            else:
                template = getattr(module, class_name)()

        return template


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""FedProx Keras optimizer module."""
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.python.ops import standard_ops


@keras.utils.register_keras_serializable()
class FedProxOptimizer(keras.optimizers.Optimizer):
    """FedProx optimizer.

    Paper: https://arxiv.org/pdf/1812.06127.pdf
    """

    def __init__(self, learning_rate=0.01, mu=0.01, name='FedProxOptimizer', **kwargs):
        """Initialize."""
        super().__init__(name=name, **kwargs)

        self._set_hyper('learning_rate', learning_rate)
        self._set_hyper('mu', mu)

        self._lr_t = None
        self._mu_t = None

    def _prepare(self, var_list):
        self._lr_t = tf.convert_to_tensor(self._get_hyper('learning_rate'), name='lr')
        self._mu_t = tf.convert_to_tensor(self._get_hyper('mu'), name='mu')

    def _create_slots(self, var_list):
        for v in var_list:
            self.add_slot(v, 'vstar')

    def _resource_apply_dense(self, grad, var):
        lr_t = tf.cast(self._lr_t, var.dtype.base_dtype)
        mu_t = tf.cast(self._mu_t, var.dtype.base_dtype)
        vstar = self.get_slot(var, 'vstar')

        var_update = var.assign_sub(lr_t * (grad + mu_t * (var - vstar)))

        return tf.group(*[var_update, ])

    def _apply_sparse_shared(self, grad, var, indices, scatter_add):
        lr_t = tf.cast(self._lr_t, var.dtype.base_dtype)
        mu_t = tf.cast(self._mu_t, var.dtype.base_dtype)
        vstar = self.get_slot(var, 'vstar')
        v_diff = vstar.assign(mu_t * (var - vstar), use_locking=self._use_locking)

        with tf.control_dependencies([v_diff]):
            scaled_grad = scatter_add(vstar, indices, grad)
        var_update = var.assign_sub(lr_t * scaled_grad)

        return tf.group(*[var_update, ])

    def _resource_apply_sparse(self, grad, var):
        return self._apply_sparse_shared(
            grad.values, var, grad.indices,
            lambda x, i, v: standard_ops.scatter_add(x, i, v))

    def get_config(self):
        """Return the config of the optimizer.

        An optimizer config is a Python dictionary (serializable)
        containing the configuration of an optimizer.
        The same optimizer can be reinstantiated later
        (without any saved state) from this configuration.

        Returns:
            Python dictionary.
        """
        base_config = super(FedProxOptimizer, self).get_config()
        return {
            **base_config,
            'lr': self._serialize_hyperparameter('learning_rate'),
            'mu': self._serialize_hyperparameter('mu')
        }



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Director server."""

import asyncio
import logging
import uuid
from pathlib import Path
from typing import Optional
from typing import Union

from google.protobuf.json_format import MessageToDict
from google.protobuf.json_format import ParseDict
from grpc import aio
from grpc import ssl_server_credentials

from openfl.pipelines import GenericPipeline
from openfl.protocols import base_pb2
from openfl.protocols import director_pb2
from openfl.protocols import director_pb2_grpc
from openfl.protocols.utils import construct_model_proto
from openfl.protocols.utils import deconstruct_model_proto
from openfl.protocols.utils import get_headers
from .grpc_channel_options import channel_options

logger = logging.getLogger(__name__)

CLIENT_ID_DEFAULT = '__default__'


class DirectorGRPCServer(director_pb2_grpc.DirectorServicer):
    """Director transport class."""

    def __init__(
            self, *,
            director_cls,
            tls: bool = True,
            root_certificate: Optional[Union[Path, str]] = None,
            private_key: Optional[Union[Path, str]] = None,
            certificate: Optional[Union[Path, str]] = None,
            listen_host: str = '[::]',
            listen_port: int = 50051,
            nn=True,
            **kwargs,
    ) -> None:
        """Initialize a director object."""
        # TODO: add working directory
        super().__init__()

        self.listen_uri = f'{listen_host}:{listen_port}'
        self.tls = tls
        self.root_certificate = None
        self.private_key = None
        self.certificate = None
        self._fill_certs(root_certificate, private_key, certificate)
        self.server = None
        self.root_dir = Path.cwd()
        self.director = director_cls(
            tls=self.tls,
            root_certificate=self.root_certificate,
            private_key=self.private_key,
            certificate=self.certificate,
            **kwargs
        )
        self.nn = nn

    def _fill_certs(self, root_certificate, private_key, certificate):
        """Fill certificates."""
        if self.tls:
            if not (root_certificate and private_key and certificate):
                raise Exception('No certificates provided')
            self.root_certificate = Path(root_certificate).absolute()
            self.private_key = Path(private_key).absolute()
            self.certificate = Path(certificate).absolute()

    def get_caller(self, context):
        """
        Get caller name from context.

            if tls == True: get caller name from auth_context
            if tls == False: get caller name from context header 'client_id'
        """
        if self.tls:
            return context.auth_context()['x509_common_name'][0].decode('utf-8')
        headers = get_headers(context)
        client_id = headers.get('client_id', CLIENT_ID_DEFAULT)
        return client_id

    def start(self):
        """Launch the director GRPC server."""
        loop = asyncio.get_event_loop()  # TODO: refactor after end of support for python3.6
        loop.create_task(self.director.start_experiment_execution_loop())
        loop.run_until_complete(self._run_server())

    async def _run_server(self):
        self.server = aio.server(options=channel_options)
        director_pb2_grpc.add_DirectorServicer_to_server(self, self.server)

        if not self.tls:
            self.server.add_insecure_port(self.listen_uri)
        else:
            with open(self.private_key, 'rb') as f:
                private_key_b = f.read()
            with open(self.certificate, 'rb') as f:
                certificate_b = f.read()
            with open(self.root_certificate, 'rb') as f:
                root_certificate_b = f.read()
            server_credentials = ssl_server_credentials(
                ((private_key_b, certificate_b),),
                root_certificates=root_certificate_b,
                require_client_auth=True
            )
            self.server.add_secure_port(self.listen_uri, server_credentials)
        logger.info(f'Starting server on {self.listen_uri}')
        await self.server.start()
        await self.server.wait_for_termination()

    async def UpdateShardInfo(self, request, context):  # NOQA:N802
        """Receive acknowledge shard info."""
        logger.info(f'UpdateShardInfo request has got: {request.shard_info}')
        dict_shard_info = MessageToDict(
            request.shard_info,
            preserving_proto_field_name=True
        )
        is_accepted = self.director.acknowledge_shard(dict_shard_info)
        reply = director_pb2.UpdateShardInfoResponse(accepted=is_accepted)

        return reply

    async def SetNewExperiment(self, stream, context):  # NOQA:N802
        """Request to set new experiment."""
        logger.info(f'SetNewExperiment request has got {stream}')
        # TODO: add streaming reader
        data_file_path = self.root_dir / str(uuid.uuid4())
        with open(data_file_path, 'wb') as data_file:
            async for request in stream:
                if request.experiment_data.size == len(request.experiment_data.npbytes):
                    data_file.write(request.experiment_data.npbytes)
                else:
                    raise Exception('Bad request')

        tensor_dict = None
        if request.model_proto:
            # TODO: the pipeline class should not be fixed
            tensor_dict, _ = deconstruct_model_proto(request.model_proto, GenericPipeline(nn=self.nn))

        caller = self.get_caller(context)

        is_accepted = await self.director.set_new_experiment(
            experiment_name=request.name,
            sender_name=caller,
            tensor_dict=tensor_dict,
            collaborator_names=request.collaborator_names,
            experiment_archive_path=data_file_path
        )

        logger.info('Send response')
        return director_pb2.SetNewExperimentResponse(accepted=is_accepted)

    async def GetTrainedModel(self, request, context):  # NOQA:N802
        """RPC for retrieving trained models."""
        logger.info('Request GetTrainedModel has got!')

        if request.model_type == director_pb2.GetTrainedModelRequest.BEST_MODEL:
            model_type = 'best'
        elif request.model_type == director_pb2.GetTrainedModelRequest.LAST_MODEL:
            model_type = 'last'
        else:
            logger.error('Incorrect model type')
            return director_pb2.TrainedModelResponse()

        caller = self.get_caller(context)

        trained_model_dict = self.director.get_trained_model(
            experiment_name=request.experiment_name,
            caller=caller,
            model_type=model_type
        )

        if trained_model_dict is None:
            return director_pb2.TrainedModelResponse()

        model_proto = construct_model_proto(trained_model_dict, 0, GenericPipeline())

        return director_pb2.TrainedModelResponse(model_proto=model_proto)

    async def GetExperimentData(self, request, context):  # NOQA:N802
        """Receive experiment data."""
        # TODO: add size filling
        # TODO: add experiment name field
        # TODO: rename npbytes to data
        data_file_path = self.director.get_experiment_data(request.experiment_name)
        max_buffer_size = (2 * 1024 * 1024)
        with open(data_file_path, 'rb') as df:
            while True:
                data = df.read(max_buffer_size)
                if len(data) == 0:
                    break
                yield director_pb2.ExperimentData(size=len(data), npbytes=data)

    async def WaitExperiment(self, request_iterator, context):  # NOQA:N802
        """Request for wait an experiment."""
        logger.info('Request WaitExperiment has got!')
        async for msg in request_iterator:
            logger.info(msg)
            experiment_name = await self.director.wait_experiment(msg.collaborator_name)
            logger.info(f'Experiment {experiment_name} was prepared')

            yield director_pb2.WaitExperimentResponse(experiment_name=experiment_name)

    async def GetDatasetInfo(self, request, context):  # NOQA:N802
        """Request the info about target and sample shapes in the dataset."""
        logger.info('Request GetDatasetInfo has got!')

        sample_shape, target_shape = self.director.get_dataset_info()
        shard_info = director_pb2.ShardInfo(
            sample_shape=sample_shape,
            target_shape=target_shape
        )
        resp = director_pb2.GetDatasetInfoResponse(shard_info=shard_info)
        return resp

    async def GetMetricStream(self, request, context):  # NOQA:N802
        """Request to stream metrics from the aggregator to frontend."""
        logger.info(f'Request GetMetricStream for {request.experiment_name} experiment has got!')

        caller = self.get_caller(context)
        async for metric_dict in self.director.stream_metrics(
                experiment_name=request.experiment_name, caller=caller
        ):
            if metric_dict is None:
                await asyncio.sleep(1)
                continue
            yield director_pb2.GetMetricStreamResponse(**metric_dict)

    async def RemoveExperimentData(self, request, context):  # NOQA:N802
        """Remove experiment data RPC."""
        response = director_pb2.RemoveExperimentResponse(acknowledgement=False)
        caller = self.get_caller(context)
        self.director.remove_experiment_data(
            experiment_name=request.experiment_name,
            caller=caller,
        )

        response.acknowledgement = True
        return response

    async def SetExperimentFailed(self, request, context):  # NOQA:N802
        """Set the experiment failed."""
        response = director_pb2.SetExperimentFailedResponse()
        if self.get_caller(context) != CLIENT_ID_DEFAULT:
            return response
        logger.error(f'Collaborator {request.collaborator_name} was failed with error code:'
                     f' {request.error_code}, error_description: {request.error_description}'
                     f'Stopping experiment.')
        self.director.set_experiment_failed(
            experiment_name=request.experiment_name,
            collaborator_name=request.collaborator_name
        )

        return response

    async def UpdateEnvoyStatus(self, request, context):  # NOQA:N802
        """Accept health check from envoy."""
        logger.debug(f'Request UpdateEnvoyStatus has got: {request}')
        cuda_devices_info = [
            MessageToDict(message, preserving_proto_field_name=True)
            for message in request.cuda_devices
        ]
        health_check_period = self.director.update_envoy_status(
            envoy_name=request.name,
            is_experiment_running=request.is_experiment_running,
            cuda_devices_status=cuda_devices_info
        )
        resp = director_pb2.UpdateEnvoyStatusResponse()
        resp.health_check_period.seconds = health_check_period

        return resp

    async def GetEnvoys(self, request, context):  # NOQA:N802
        """Get a status information about envoys."""
        envoy_infos = self.director.get_envoys()
        envoy_statuses = []
        for envoy_info in envoy_infos:
            envoy_info_message = director_pb2.EnvoyInfo(
                shard_info=ParseDict(
                    envoy_info['shard_info'], director_pb2.ShardInfo(),
                    ignore_unknown_fields=True),
                is_online=envoy_info['is_online'],
                is_experiment_running=envoy_info['is_experiment_running'])
            envoy_info_message.valid_duration.seconds = envoy_info['valid_duration']
            envoy_info_message.last_updated.seconds = int(envoy_info['last_updated'])

            envoy_statuses.append(envoy_info_message)

        return director_pb2.GetEnvoysResponse(envoy_infos=envoy_statuses)

    async def GetExperimentsList(self, request, context):  # NOQA:N802
        """Get list of experiments description."""
        caller = self.get_caller(context)
        experiments = self.director.get_experiments_list(caller)
        experiment_list = [
            director_pb2.ExperimentListItem(**exp)
            for exp in experiments
        ]
        return director_pb2.GetExperimentsListResponse(
            experiments=experiment_list
        )

    async def GetExperimentDescription(self, request, context):  # NOQA:N802
        """Get an experiment description."""
        caller = self.get_caller(context)
        experiment = self.director.get_experiment_description(caller, request.name)
        models_statuses = [
            base_pb2.DownloadStatus(
                name=ms['name'],
                status=ms['status']
            )
            for ms in experiment['download_statuses']['models']
        ]
        logs_statuses = [
            base_pb2.DownloadStatus(
                name=ls['name'],
                status=ls['status']
            )
            for ls in experiment['download_statuses']['logs']
        ]
        download_statuses = base_pb2.DownloadStatuses(
            models=models_statuses,
            logs=logs_statuses,
        )
        collaborators = [
            base_pb2.CollaboratorDescription(
                name=col['name'],
                status=col['status'],
                progress=col['progress'],
                round=col['round'],
                current_task=col['current_task'],
                next_task=col['next_task']
            )
            for col in experiment['collaborators']
        ]
        tasks = [
            base_pb2.TaskDescription(
                name=task['name'],
                description=task['description']
            )
            for task in experiment['tasks']
        ]

        return director_pb2.GetExperimentDescriptionResponse(
            experiment=base_pb2.ExperimentDescription(
                name=experiment['name'],
                status=experiment['status'],
                progress=experiment['progress'],
                current_round=experiment['current_round'],
                total_rounds=experiment['total_rounds'],
                download_statuses=download_statuses,
                collaborators=collaborators,
                tasks=tasks,
            ),
        )



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Generic check functions."""
import os


def get_credentials(folder_path):
    """Get credentials from folder by template."""
    root_ca, key, cert = None, None, None
    if os.path.exists(folder_path):
        for f in os.listdir(folder_path):
            if '.key' in f:
                key = folder_path + '/' + f
            if '.crt' in f and 'root_ca' not in f:
                cert = folder_path + '/' + f
            if 'root_ca' in f:
                root_ca = folder_path + '/' + f
    return root_ca, key, cert



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Generic check functions."""


def check_type(obj, expected_type, logger):
    """Assert `obj` is of `expected_type` type."""
    if not isinstance(obj, expected_type):
        exception = TypeError(f'Expected type {type(obj)}, got type {str(expected_type)}')
        logger.exception(repr(exception))
        raise exception


def check_equal(x, y, logger):
    """Assert `x` and `y` are equal."""
    if x != y:
        exception = ValueError(f'{x} != {y}')
        logger.exception(repr(exception))
        raise exception


def check_not_equal(x, y, logger, name='None provided'):
    """Assert `x` and `y` are not equal."""
    if x == y:
        exception = ValueError(f'Name {name}. Expected inequality, but {x} == {y}')
        logger.exception(repr(exception))
        raise exception


def check_is_in(element, _list, logger):
    """Assert `element` is in `_list` collection."""
    if element not in _list:
        exception = ValueError(f'Expected sequence membership, but {element} is not in {_list}')
        logger.exception(repr(exception))
        raise exception


def check_not_in(element, _list, logger):
    """Assert `element` is not in `_list` collection."""
    if element in _list:
        exception = ValueError(f'Expected not in sequence, but {element} is in {_list}')
        logger.exception(repr(exception))
        raise exception



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""NoCompressionPipeline module."""

from .pipeline import Float32NumpyArrayToBytes
from .pipeline import TransformationPipeline


class NoCompressionPipeline(TransformationPipeline):
    """The data pipeline without any compression."""

    def __init__(self, **kwargs):
        """Initialize."""
        super(NoCompressionPipeline, self).__init__(
            transformers=[Float32NumpyArrayToBytes()], **kwargs)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""CIFAR10 Shard Descriptor (using `TFDS` API)"""
import jax
import jax.numpy as jnp
import logging
import tensorflow as tf
import tensorflow_datasets as tfds

from typing import List, Tuple
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)

class CIFAR10ShardDescriptor(ShardDescriptor):
    """
    CIFAR10 Shard Descriptor

    This example is based on `tfds` data loader.
    Note that the ingestion of any model/task requires an iterable dataloader.
    Hence, it is possible to utilize these pipelines without explicit need of a
    new interface.
    """

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            **kwargs
    ) -> None:
        """Download/Prepare CIFAR10 dataset"""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))

        # Load dataset
        train_ds, valid_ds = self._download_and_prepare_dataset(self.rank, self.worldsize)

        # Set attributes
        self._sample_shape = train_ds['image'].shape[1:]
        self._target_shape = tf.expand_dims(train_ds['label'], -1).shape[1:]

        self.splits = {
            'train': train_ds,
            'valid': valid_ds
        }

    def _download_and_prepare_dataset(self, rank: int, worldsize: int) -> Tuple[dict]:
        """
        Download, Cache CIFAR10 and prepare `tfds` builder.

        Provide `rank` and `worldsize` to virtually split dataset across shards uniquely for each client
        for simulation purposes.

        Returns:
        Tuple (train_dict, test_dict) of dictionary with JAX DeviceArray (image and label)
        dict['image'] -> DeviceArray float32
        dict['label'] -> DeviceArray int32
        {'image' : DeviceArray(...), 'label' : DeviceArray(...)}
        
        """
        
        dataset_builder = tfds.builder('cifar10')
        dataset_builder.download_and_prepare()

        datasets, dataset_info = dataset_builder.as_dataset(), dataset_builder.info

        train_shard_size = int(len(datasets['train']) / worldsize)
        test_shard_size = int(len(datasets['test']) / worldsize)

        self.train_segment = f'train[{train_shard_size * (rank - 1)}:{train_shard_size * rank}]'
        self.test_segment = f'test[{test_shard_size * (rank - 1)}:{test_shard_size * rank}]'

        train_ds = tfds.as_numpy(dataset_builder.as_dataset(split=self.train_segment, batch_size=-1))
        test_ds = tfds.as_numpy(dataset_builder.as_dataset(split=self.test_segment, batch_size=-1))

        train_ds['image'] = jnp.float32(train_ds['image']) / 255.
        test_ds['image'] = jnp.float32(test_ds['image']) / 255.
        train_ds['label'] = jnp.int32(train_ds['label'])
        test_ds['label'] = jnp.int32(test_ds['label'])

        return train_ds, test_ds

    def get_shard_dataset_types(self) -> List[str]:
        """Get available split names"""
        return list(self.splits)

    def get_split(self, name: str) -> tf.data.Dataset:
        """Return a shard dataset by type."""
        if name not in self.splits:
            raise Exception(f'Split name `{name}` not found.'
                            f' Expected one of {list(self.splits.keys())}')
        return self.splits[name]

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        return list(map(str, self._sample_shape))

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        return list(map(str, self._target_shape))

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        n_train = len(self.splits['train']['label'])
        n_test = len(self.splits['valid']['label'])
        return (f'CIFAR10 dataset, Shard Segments {self.train_segment}/{self.test_segment}, rank/world {self.rank}/{self.worldsize}.'
                f'\n num_samples [Train/Valid]: [{n_train}/{n_test}]')


# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregation functions package."""

from .adagrad_adaptive_aggregation import AdagradAdaptiveAggregation
from .adam_adaptive_aggregation import AdamAdaptiveAggregation
from .core import AggregationFunction
from .fedcurv_weighted_average import FedCurvWeightedAverage
from .geometric_median import GeometricMedian
from .median import Median
from .weighted_average import WeightedAverage
from .yogi_adaptive_aggregation import YogiAdaptiveAggregation
from .aggregate_random_forest import AggregateRandomForest
from .aggregate_adaboost import AggregateAdaboost

__all__ = ['Median',
           'WeightedAverage',
           'GeometricMedian',
           'AdagradAdaptiveAggregation',
           'AdamAdaptiveAggregation',
           'YogiAdaptiveAggregation',
           'AggregationFunction',
           'FedCurvWeightedAverage',
           'AggregateRandomForest',
           'AggregateAdaboost']



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""MedMNIST Shard Descriptor."""

import logging
import os
from typing import Any, Dict, List, Tuple
from medmnist.info import INFO, HOMEPAGE

import numpy as np

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)


class MedMNISTShardDataset(ShardDataset):
    """MedMNIST Shard dataset class."""

    def __init__(self, x, y, data_type: str = 'train', rank: int = 1, worldsize: int = 1) -> None:
        """Initialize MedMNISTDataset."""
        self.data_type = data_type
        self.rank = rank
        self.worldsize = worldsize
        self.x = x[self.rank - 1::self.worldsize]
        self.y = y[self.rank - 1::self.worldsize]

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """Return an item by the index."""
        return self.x[index], self.y[index]

    def __len__(self) -> int:
        """Return the len of the dataset."""
        return len(self.x)


class MedMNISTShardDescriptor(ShardDescriptor):
    """MedMNIST Shard descriptor class."""

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            datapath: str = '',
            dataname: str = 'bloodmnist',
            **kwargs
    ) -> None:
        """Initialize MedMNISTShardDescriptor."""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        
        self.datapath = datapath
        self.dataset_name = dataname;
        self.info = INFO[self.dataset_name]
        
        (x_train, y_train), (x_test, y_test) = self.load_data()
        self.data_by_type = {
            'train': (x_train, y_train),
            'val': (x_test, y_test)
        }

    def get_shard_dataset_types(self) -> List[str]:
        """Get available shard dataset types."""
        return list(self.data_by_type)

    def get_dataset(self, dataset_type='train') -> MedMNISTShardDataset:
        """Return a shard dataset by type."""
        if dataset_type not in self.data_by_type:
            raise Exception(f'Wrong dataset type: {dataset_type}')
        return MedMNISTShardDataset(
            *self.data_by_type[dataset_type],
            data_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self) -> List[str] :
        """Return the sample shape info."""
        return ['28', '28', '3']

    @property
    def target_shape(self) -> List[str] :
        """Return the target shape info."""
        return ['1','1']

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'MedMNIST dataset, shard number {self.rank}'
                f' out of {self.worldsize}')

    def download_data(datapath: str = 'data/', 
                      dataname: str = 'bloodmnist', 
                      info: dict = {}) -> None:

        logger.info(f"{datapath}\n{dataname}\n{info}")
        try:
            from torchvision.datasets.utils import download_url
            download_url(url=info["url"],
                         root=datapath,
                         filename=dataname,
                         md5=info["MD5"])
        except:
            raise RuntimeError('Something went wrong when downloading! ' +
                               'Go to the homepage to download manually. ' +
                               HOMEPAGE)


    def load_data(self) -> Tuple[Tuple[Any, Any], Tuple[Any, Any]]:
        """Download prepared dataset."""
        
        dataname = self.dataset_name + '.npz'
        dataset = os.path.join(self.datapath, dataname)
        
        if not os.path.isfile(dataset):
            logger.info(f"Dataset {dataname} not found at:{self.datapath}.\n\tDownloading...")
            MedMNISTShardDescriptor.download_data(self.datapath, dataname, self.info)
            logger.info(f"DONE!")

        data = np.load(dataset)
        
        x_train = data["train_images"]
        x_test = data["test_images"]
        
        y_train=data["train_labels"]
        y_test=data["test_labels"]
        logger.info('MedMNIST data was loaded!')
        return (x_train, y_train), (x_test, y_test)
        
        



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Layers for Unet model."""

import torch
import torch.nn as nn
import torch.nn.functional as F


def soft_dice_loss(output, target):
    """Calculate loss."""
    num = target.size(0)
    m1 = output.view(num, -1)
    m2 = target.view(num, -1)
    intersection = m1 * m2
    score = 2.0 * (intersection.sum(1) + 1) / (m1.sum(1) + m2.sum(1) + 1)
    score = 1 - score.sum() / num
    return score


def soft_dice_coef(output, target):
    """Calculate soft DICE coefficient."""
    num = target.size(0)
    m1 = output.view(num, -1)
    m2 = target.view(num, -1)
    intersection = m1 * m2
    score = 2.0 * (intersection.sum(1) + 1) / (m1.sum(1) + m2.sum(1) + 1)
    return score.sum()


class DoubleConv(nn.Module):
    """Pytorch double conv class."""

    def __init__(self, in_ch, out_ch):
        """Initialize layer."""
        super(DoubleConv, self).__init__()
        self.in_ch = in_ch
        self.out_ch = out_ch
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        """Do forward pass."""
        x = self.conv(x)
        return x


class Down(nn.Module):
    """Pytorch nn module subclass."""

    def __init__(self, in_ch, out_ch):
        """Initialize layer."""
        super(Down, self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_ch, out_ch)
        )

    def forward(self, x):
        """Do forward pass."""
        x = self.mpconv(x)
        return x


class Up(nn.Module):
    """Pytorch nn module subclass."""

    def __init__(self, in_ch, out_ch, bilinear=False):
        """Initialize layer."""
        super(Up, self).__init__()
        self.in_ch = in_ch
        self.out_ch = out_ch
        if bilinear:
            self.up = nn.Upsample(
                scale_factor=2,
                mode='bilinear',
                align_corners=True
            )
        else:
            self.up = nn.ConvTranspose2d(in_ch, in_ch // 2, 2, stride=2)
        self.conv = DoubleConv(in_ch, out_ch)

    def forward(self, x1, x2):
        """Do forward pass."""
        x1 = self.up(x1)
        diff_y = x2.size()[2] - x1.size()[2]
        diff_x = x2.size()[3] - x1.size()[3]

        x1 = F.pad(
            x1,
            (diff_x // 2, diff_x - diff_x // 2, diff_y // 2, diff_y - diff_y // 2)
        )

        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Superb Shard Descriptor."""

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from datasets import load_dataset
from datasets import load_metric

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor


class SuperbShardDataset(ShardDataset):
    """SUPERB Shard dataset class."""

    def __init__(self, dataset, rank: int = 1, worldsize: int = 1) -> None:
        """Initialize Superb shard Dataset."""
        self.rank = rank
        self.worldsize = worldsize
        self.dataset = dataset
        self.x = self.dataset['audio'][self.rank - 1::self.worldsize]
        self.y = self.dataset['label'][self.rank - 1::self.worldsize]

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """Return an item by the index."""
        return self.x[index]['array'], self.y[index]

    def __len__(self) -> int:
        """Return the len of the dataset."""
        return len(self.x)


class SuperbShardDescriptor(ShardDescriptor):
    """Superb Shard descriptor class."""

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            **kwargs
    ) -> None:
        """Initialize SuperbShardDescriptor."""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        print('rank and worldsize', self.rank, self.worldsize)
        train_set, val_set, test_set = self.download_data()
        self.data_by_type = {
            'train': train_set,
            'val': val_set,
            'test': test_set
        }

    def get_shard_dataset_types(self) -> List[Dict[str, Any]]:
        """Get available shard dataset types."""
        return list(self.data_by_type)

    def get_dataset(self, dataset_type: str = 'train') -> SuperbShardDataset:
        """Return a shard dataset by type."""
        if dataset_type not in self.data_by_type:
            raise Exception(f'Wrong dataset type: {dataset_type}')
        return SuperbShardDataset(
            self.data_by_type[dataset_type],
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        return ['1']

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        return ['1']

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Superb dataset, shard number {self.rank}'
                f' out of {self.worldsize}')

    def download_data(self) -> Tuple[Tuple[Dict, List], Tuple[Dict, List], Tuple[Dict, List]]:
        """Download dataset."""
        dataset = load_dataset('superb', 'ks')
        metric = load_metric('accuracy') # noqa

        # Train data
        train_set = dataset['train']

        # Validation data
        val_set = dataset['validation']

        # Test data
        test_set = dataset['test']

        print('Superb data was loaded!')
        return train_set, val_set, test_set



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl common object types."""

from abc import ABCMeta
from collections import namedtuple

TensorKey = namedtuple('TensorKey', ['tensor_name', 'origin', 'round_number', 'report', 'tags'])
TaskResultKey = namedtuple('TaskResultKey', ['task_name', 'owner', 'round_number'])

Metric = namedtuple('Metric', ['name', 'value'])
LocalTensor = namedtuple('LocalTensor', ['col_name', 'tensor', 'weight'])


class SingletonABCMeta(ABCMeta):
    """Metaclass for singleton instances."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Use the singleton instance if it has already been created."""
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Module with auxiliary CLI helper functions."""

from itertools import islice
from os import environ
from os import stat
from pathlib import Path
from sys import argv

from click import echo
from click import style
from yaml import FullLoader
from yaml import load

FX = argv[0]

SITEPACKS = Path(__file__).parent.parent.parent
WORKSPACE = SITEPACKS / 'openfl-workspace'
TUTORIALS = SITEPACKS / 'openfl-tutorials'
OPENFL_USERDIR = Path.home() / '.openfl'
CERT_DIR = Path('cert').absolute()


def pretty(o):
    """Pretty-print the dictionary given."""
    m = max(map(len, o.keys()))

    for k, v in o.items():
        echo(style(f'{k:<{m}} : ', fg='blue') + style(f'{v}', fg='cyan'))


def tree(path):
    """Print current directory file tree."""
    echo(f'+ {path}')

    for path in sorted(path.rglob('*')):

        depth = len(path.relative_to(path).parts)
        space = '    ' * depth

        if path.is_file():
            echo(f'{space}f {path.name}')
        else:
            echo(f'{space}d {path.name}')


def print_tree(dir_path: Path, level: int = -1,
               limit_to_directories: bool = False,
               length_limit: int = 1000):
    """Given a directory Path object print a visual tree structure."""
    space = '    '
    branch = '│   '
    tee = '├── '
    last = '└── '

    echo('\nNew workspace directory structure:')

    dir_path = Path(dir_path)  # accept string coerceable to Path
    files = 0
    directories = 0

    def inner(dir_path: Path, prefix: str = '', level=-1):
        nonlocal files, directories
        if not level:
            return  # 0, stop iterating
        if limit_to_directories:
            contents = [d for d in dir_path.iterdir() if d.is_dir()]
        else:
            contents = list(dir_path.iterdir())
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                yield prefix + pointer + path.name
                directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension,
                                 level=level - 1)
            elif not limit_to_directories:
                yield prefix + pointer + path.name
                files += 1

    echo(dir_path.name)
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        echo(line)
    if next(iterator, None):
        echo(f'... length_limit, {length_limit}, reached, counted:')
    echo(f'\n{directories} directories' + (f', {files} files' if files else ''))


def copytree(src, dst, symlinks=False, ignore=None,
             ignore_dangling_symlinks=False, dirs_exist_ok=False):
    """From Python 3.8 'shutil' which include 'dirs_exist_ok' option."""
    import os
    import shutil

    with os.scandir(src) as itr:
        entries = list(itr)

    copy_function = shutil.copy2

    def _copytree():

        if ignore is not None:
            ignored_names = ignore(os.fspath(src), [x.name for x in entries])
        else:
            ignored_names = set()

        os.makedirs(dst, exist_ok=dirs_exist_ok)
        errors = []
        use_srcentry = copy_function is shutil.copy2 or copy_function is shutil.copy

        for srcentry in entries:
            if srcentry.name in ignored_names:
                continue
            srcname = os.path.join(src, srcentry.name)
            dstname = os.path.join(dst, srcentry.name)
            srcobj = srcentry if use_srcentry else srcname
            try:
                is_symlink = srcentry.is_symlink()
                if is_symlink and os.name == 'nt':
                    lstat = srcentry.stat(follow_symlinks=False)
                    if lstat.st_reparse_tag == stat.IO_REPARSE_TAG_MOUNT_POINT:
                        is_symlink = False
                if is_symlink:
                    linkto = os.readlink(srcname)
                    if symlinks:
                        os.symlink(linkto, dstname)
                        shutil.copystat(srcobj, dstname,
                                        follow_symlinks=not symlinks)
                    else:
                        if (not os.path.exists(linkto)
                                and ignore_dangling_symlinks):
                            continue
                        if srcentry.is_dir():
                            copytree(srcobj, dstname, symlinks, ignore,
                                     dirs_exist_ok=dirs_exist_ok)
                        else:
                            copy_function(srcobj, dstname)
                elif srcentry.is_dir():
                    copytree(srcobj, dstname, symlinks, ignore,
                             dirs_exist_ok=dirs_exist_ok)
                else:
                    copy_function(srcobj, dstname)
            except OSError as why:
                errors.append((srcname, dstname, str(why)))
            except Exception as err:
                errors.extend(err.args[0])
        try:
            shutil.copystat(src, dst)
        except OSError as why:
            if getattr(why, 'winerror', None) is None:
                errors.append((src, dst, str(why)))
        if errors:
            raise Exception(errors)
        return dst

    return _copytree()


def get_workspace_parameter(name):
    """Get a parameter from the workspace config file (.workspace)."""
    # Update the .workspace file to show the current workspace plan
    workspace_file = '.workspace'

    with open(workspace_file, 'r') as f:
        doc = load(f, Loader=FullLoader)

    if not doc:  # YAML is not correctly formatted
        doc = {}  # Create empty dictionary

    if name not in doc.keys() or not doc[name]:  # List doesn't exist
        return ''
    else:
        return doc[name]


def check_varenv(env: str = '', args: dict = None):
    """Update "args" (dictionary) with <env: env_value> if env has a defined value in the host."""
    if args is None:
        args = {}
    env_val = environ.get(env)
    if env and (env_val is not None):
        args[env] = env_val

    return args


def get_fx_path(curr_path=''):
    """Return the absolute path to fx binary."""
    import re
    import os

    match = re.search('lib', curr_path)
    idx = match.end()
    path_prefix = curr_path[0:idx]
    bin_path = re.sub('lib', 'bin', path_prefix)
    fx_path = os.path.join(bin_path, 'fx')

    return fx_path


def remove_line_from_file(pkg, filename):
    """Remove line that contains `pkg` from the `filename` file."""
    with open(filename, 'r+') as f:
        d = f.readlines()
        f.seek(0)
        for i in d:
            if pkg not in i:
                f.write(i)
        f.truncate()


def replace_line_in_file(line, line_num_to_replace, filename):
    """Replace line at `line_num_to_replace` with `line`."""
    with open(filename, 'r+') as f:
        d = f.readlines()
        f.seek(0)
        for idx, i in enumerate(d):
            if idx == line_num_to_replace:
                f.write(line)
            else:
                f.write(i)
        f.truncate()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""CUDA Device monitor plugin module."""

from .device_monitor import DeviceMonitor


class CUDADeviceMonitor(DeviceMonitor):
    """CUDA Device monitor plugin."""

    def get_driver_version(self) -> str:
        """Get CUDA driver version."""
        raise NotImplementedError

    def get_device_memory_total(self, index: int) -> int:
        """Get total memory available on the device."""
        raise NotImplementedError

    def get_device_memory_utilized(self, index: int) -> int:
        """Get utilized memory on the device."""
        raise NotImplementedError

    def get_device_utilization(self, index: int) -> str:
        """
        Get device utilization.

        It is just a general method that returns a string that may be shown to the frontend user.
        """
        raise NotImplementedError

    def get_device_name(self, index: int) -> str:
        """Get device name."""
        raise NotImplementedError

    def get_cuda_version(self) -> str:
        """Get CUDA driver version."""
        raise NotImplementedError



import torch.nn as nn
import numpy as np
from .batchnorm import SynchronizedBatchNorm3d, SynchronizedBatchNorm2d

class Transform3D:

    def __init__(self, mul=None):
        self.mul = mul

    def __call__(self, voxel):
   
        if self.mul == '0.5':
            voxel = voxel * 0.5
        elif self.mul == 'random':
            voxel = voxel * np.random.uniform()
        
        return voxel.astype(np.float32)


def model_to_syncbn(model):
    preserve_state_dict = model.state_dict()
    _convert_module_from_bn_to_syncbn(model)
    model.load_state_dict(preserve_state_dict)
    return model


def _convert_module_from_bn_to_syncbn(module):
    for child_name, child in module.named_children(): 
        if hasattr(nn, child.__class__.__name__) and \
            'batchnorm' in child.__class__.__name__.lower():
            TargetClass = globals()['Synchronized'+child.__class__.__name__]
            arguments = TargetClass.__init__.__code__.co_varnames[1:]
            kwargs = {k: getattr(child, k) for k in arguments}
            setattr(module, child_name, TargetClass(**kwargs))
        else:
            _convert_module_from_bn_to_syncbn(child)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Shard descriptor for text."""

import re
import urllib.request
import zipfile
from pathlib import Path

import gdown
import numpy as np
import pandas as pd

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor


class NextWordShardDataset(ShardDataset):
    """Shard Dataset for text."""

    def __init__(self, X, y):
        """Initialize NextWordShardDataset."""
        self.X = X
        self.y = y

    def __len__(self):
        """Count number of sequences."""
        return len(self.X)

    def __getitem__(self, index: int):
        """Return an item by the index."""
        return self.X[index], self.y[index]


class NextWordShardDescriptor(ShardDescriptor):
    """Data is any text."""

    def __init__(self, title: str = '', author: str = '') -> None:
        """Initialize NextWordShardDescriptor."""
        super().__init__()

        self.title = title
        self.author = author

        dataset_dir = self.download_data(title)
        data = self.load_data(dataset_dir)  # list of words
        self.X, self.y = self.get_sequences(data)

    def get_dataset(self, dataset_type='train', train_val_split=0.8):
        """Return a dataset by type."""
        train_size = round(len(self.X) * train_val_split)
        if dataset_type == 'train':
            X = self.X[:train_size]
            y = self.y[:train_size]
        elif dataset_type == 'val':
            X = self.X[train_size:]
            y = self.y[train_size:]
        else:
            raise Exception(f'Wrong dataset type: {dataset_type}.'
                            f'Choose from the list: [train, val]')
        return NextWordShardDataset(X, y)

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        length, n_gram, vector_size = self.X.shape
        return [str(n_gram), str(vector_size)]  # three vectors

    @property
    def target_shape(self):
        """Return the target shape info."""
        length, vocab_size = self.y.shape
        return [str(vocab_size)]  # row at one-hot matrix with n = vocab_size

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return f'Dataset from {self.title} by {self.author}'

    @staticmethod
    def load_data(path):
        """Load text file, return list of words."""
        file = open(path, 'r', encoding='utf8').read()
        data = re.findall(r'[a-z]+', file.lower())
        return data

    @staticmethod
    def get_sequences(data):
        """Transform words to sequences, for X transform to vectors as well."""
        # spacy en_core_web_sm vocab_size = 10719, vector_size = 96
        x_seq = []
        y_seq = []

        # created with make_vocab.py from
        # https://gist.github.com/katerina-merkulova/e351b11c67832034b49652835b14adb0
        NextWordShardDescriptor.download_vectors()
        vectors = pd.read_feather('keyed_vectors.feather')
        vectors.set_index('index', inplace=True)

        for i in range(len(data) - 3):
            x = data[i:i + 3]  # make 3-grams
            y = data[i + 3]
            cur_x = [vectors.vector[word] for word in x if word in vectors.index]
            if len(cur_x) == 3 and y in vectors.index:
                x_seq.append(cur_x)
                y_seq.append(vectors.index.get_loc(y))

        x_seq = np.array(x_seq)
        y_seq = np.array(y_seq)
        y = np.zeros((y_seq.size, 10719))
        y[np.arange(y_seq.size), y_seq] = 1
        return x_seq, y

    @staticmethod
    def download_data(title):
        """Download text by title form Github Gist."""
        url = ('https://gist.githubusercontent.com/katerina-merkulova/e351b11c67832034b49652835b'
               '14adb0/raw/5b6667c3a2e1266f3d9125510069d23d8f24dc73/' + title.replace(' ', '_')
               + '.txt')
        filepath = Path.cwd() / f'{title}.txt'
        if not filepath.exists():
            with urllib.request.urlopen(url) as response:
                content = response.read().decode('utf-8')
            with open(filepath, 'w', encoding='utf-8') as file:
                file.write(content)
        return filepath

    @staticmethod
    def download_vectors():
        """Download vectors."""
        if Path('keyed_vectors.feather').exists():
            return None

        output = 'keyed_vectors.zip'
        if not Path(output).exists():
            url = 'https://drive.google.com/uc?id=1QfidtkJ9qxzNLs1pgXoY_hqnBjsDI_2i'
            gdown.download(url, output, quiet=False)

        with zipfile.ZipFile(output, 'r') as zip_ref:
            zip_ref.extractall(Path.cwd())

        Path(output).unlink()  # remove zip



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""CIFAR10 Shard Descriptor (using `tf.data.Dataset` API)"""
import logging
from typing import List, Tuple

import tensorflow as tf

from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)


class CIFAR10ShardDescriptor(ShardDescriptor):
    """
    CIFAR10 Shard Descriptor

    This example is based on `tf.data.Dataset` pipelines.
    Note that the ingestion of any model/task requires an iterable dataloader.
    Hence, it is possible to utilize these pipelines without explicit need of a
    new interface.
    """

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            **kwargs
    ):
        """Download/Prepare CIFAR10 dataset"""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))

        # Load dataset
        train_ds, valid_ds = self._download_and_prepare_dataset(rank=self.rank, worldsize=self.worldsize)

        # Set attributes
        self._sample_shape = train_ds.element_spec[0].shape
        self._target_shape = train_ds.element_spec[1].shape

        self.splits = {
            'train': train_ds,
            'valid': valid_ds
        }

    @staticmethod
    def _download_and_prepare_dataset(rank: int, worldsize: int) -> Tuple[tf.data.Dataset]:
        """
        Load CIFAR10 as `tf.data.Dataset`.

        Provide `rank` and `worldsize` to auto-split uniquely for each client
        for simulation purposes.
        """
        (x_train, y_train), (x_test, y_test) = tf.keras.datasets.cifar10.load_data()

        # Split
        x_train, y_train = x_train[rank-1::worldsize], y_train[rank-1::worldsize]
        x_test, y_test = x_test[rank-1::worldsize], y_test[rank-1::worldsize]

        train_ds = tf.data.Dataset.from_tensor_slices((x_train, y_train))
        test_ds = tf.data.Dataset.from_tensor_slices((x_test, y_test))
        return train_ds, test_ds

    def get_shard_dataset_types(self) -> List[str]:
        """Get available split names"""
        return list(self.splits)

    def get_split(self, name: str) -> tf.data.Dataset:
        """Return a shard dataset by type."""
        if name not in self.splits:
            raise Exception(f'Split name `{name}` not found.'
                            f' Expected one of {list(self.splits.keys())}')
        return self.splits[name]

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        return list(map(str, self._sample_shape))

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        return list(map(str, self._target_shape))

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        n_train = len(self.splits['train'])
        n_test = len(self.splits['valid'])
        return (f'CIFAR10 dataset, shard number {self.rank}/{self.worldsize}.'
                f'\nSamples [Train/Valid]: [{n_train}/{n_test}]')


# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Cryptography CA utilities."""

import datetime
import uuid
from typing import Tuple

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509.base import Certificate
from cryptography.x509.base import CertificateSigningRequest
from cryptography.x509.extensions import ExtensionNotFound
from cryptography.x509.name import Name
from cryptography.x509.oid import ExtensionOID
from cryptography.x509.oid import NameOID


def generate_root_cert(days_to_expiration: int = 365) -> Tuple[RSAPrivateKey, Certificate]:
    """Generate_root_certificate."""
    now = datetime.datetime.utcnow()
    expiration_delta = days_to_expiration * datetime.timedelta(1, 0, 0)

    # Generate private key
    root_private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=3072,
        backend=default_backend()
    )

    # Generate public key
    root_public_key = root_private_key.public_key()
    builder = x509.CertificateBuilder()
    subject = x509.Name([
        x509.NameAttribute(NameOID.DOMAIN_COMPONENT, u'org'),
        x509.NameAttribute(NameOID.DOMAIN_COMPONENT, u'simple'),
        x509.NameAttribute(NameOID.COMMON_NAME, u'Simple Root CA'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u'Simple Inc'),
        x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, u'Simple Root CA'),
    ])
    issuer = subject
    builder = builder.subject_name(subject)
    builder = builder.issuer_name(issuer)

    builder = builder.not_valid_before(now)
    builder = builder.not_valid_after(now + expiration_delta)
    builder = builder.serial_number(int(uuid.uuid4()))
    builder = builder.public_key(root_public_key)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=True, path_length=None), critical=True,
    )

    # Sign the CSR
    certificate = builder.sign(
        private_key=root_private_key, algorithm=hashes.SHA384(),
        backend=default_backend()
    )

    return root_private_key, certificate


def generate_signing_csr() -> Tuple[RSAPrivateKey, CertificateSigningRequest]:
    """Generate signing CSR."""
    # Generate private key
    signing_private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=3072,
        backend=default_backend()
    )

    builder = x509.CertificateSigningRequestBuilder()
    subject = x509.Name([
        x509.NameAttribute(NameOID.DOMAIN_COMPONENT, u'org'),
        x509.NameAttribute(NameOID.DOMAIN_COMPONENT, u'simple'),
        x509.NameAttribute(NameOID.COMMON_NAME, u'Simple Signing CA'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u'Simple Inc'),
        x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, u'Simple Signing CA'),
    ])
    builder = builder.subject_name(subject)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=True, path_length=None), critical=True,
    )

    # Sign the CSR
    csr = builder.sign(
        private_key=signing_private_key, algorithm=hashes.SHA384(),
        backend=default_backend()
    )

    return signing_private_key, csr


def sign_certificate(csr: CertificateSigningRequest, issuer_private_key: RSAPrivateKey,
                     issuer_name: Name, days_to_expiration: int = 365,
                     ca: bool = False) -> Certificate:
    """
    Sign the incoming CSR request.

    Args:
        csr                : Certificate Signing Request object
        issuer_private_key : Root CA private key if the request is for the signing
                             CA; Signing CA private key otherwise
        issuer_name        : x509 Name
        days_to_expiration : int (365 days by default)
        ca                 : Is this a certificate authority
    """
    now = datetime.datetime.utcnow()
    expiration_delta = days_to_expiration * datetime.timedelta(1, 0, 0)

    builder = x509.CertificateBuilder()
    builder = builder.subject_name(csr.subject)
    builder = builder.issuer_name(issuer_name)
    builder = builder.not_valid_before(now)
    builder = builder.not_valid_after(now + expiration_delta)
    builder = builder.serial_number(int(uuid.uuid4()))
    builder = builder.public_key(csr.public_key())
    builder = builder.add_extension(
        x509.BasicConstraints(ca=ca, path_length=None), critical=True,
    )
    try:
        builder = builder.add_extension(
            csr.extensions.get_extension_for_oid(
                ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            ).value, critical=False
        )
    except ExtensionNotFound:
        pass  # Might not have alternative name

    signed_cert = builder.sign(
        private_key=issuer_private_key, algorithm=hashes.SHA384(), backend=default_backend()
    )
    return signed_cert



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Cryptography IO utilities."""

from hashlib import sha384
from pathlib import Path
from typing import Tuple

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.x509.base import Certificate
from cryptography.x509.base import CertificateSigningRequest


def read_key(path: Path) -> RSAPrivateKey:
    """
    Read private key.

    Args:
        path : Path (pathlib)

    Returns:
        private_key
    """
    with open(path, 'rb') as f:
        pem_data = f.read()

    signing_key = load_pem_private_key(pem_data, password=None)
    assert(isinstance(signing_key, rsa.RSAPrivateKey))
    return signing_key


def write_key(key: RSAPrivateKey, path: Path) -> None:
    """
    Write private key.

    Args:
        key  : RSA private key object
        path : Path (pathlib)

    """
    with open(path, 'wb') as f:
        f.write(key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ))


def read_crt(path: Path) -> Certificate:
    """
    Read signed TLS certificate.

    Args:
        path : Path (pathlib)

    Returns:
        Cryptography TLS Certificate object
    """
    with open(path, 'rb') as f:
        pem_data = f.read()

    certificate = x509.load_pem_x509_certificate(pem_data)
    assert(isinstance(certificate, x509.Certificate))
    return certificate


def write_crt(certificate: Certificate, path: Path) -> None:
    """
    Write cryptography certificate / csr.

    Args:
        certificate : cryptography csr / certificate object
        path : Path (pathlib)

    Returns:
        Cryptography TLS Certificate object
    """
    with open(path, 'wb') as f:
        f.write(certificate.public_bytes(
            encoding=serialization.Encoding.PEM,
        ))


def read_csr(path: Path) -> Tuple[CertificateSigningRequest, str]:
    """
    Read certificate signing request.

    Args:
        path : Path (pathlib)

    Returns:
        Cryptography CSR object
    """
    hasher = sha384()
    with open(path, 'rb') as f:
        pem_data = f.read()
        hasher.update(pem_data)

    csr = x509.load_pem_x509_csr(pem_data)
    assert(isinstance(csr, x509.CertificateSigningRequest))
    return csr, hasher.hexdigest()



"""Copyright (C) 2020-2021 Intel Corporation
   SPDX-License-Identifier: Apache-2.0

Licensed subject to the terms of the separately executed evaluation
license agreement between Intel Corporation and you.
"""
from logging import getLogger
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import src.dataloader_utils as dlu

from openfl.federated import KerasDataLoader

logger = getLogger(__name__)


class NLPDataLoader(KerasDataLoader):
    """NLP Dataloader template."""

    def __init__(self, collaborator_count: int, split_ratio: float,
                 num_samples: int, data_path: str, batch_size: int, **kwargs) -> None:
        """Instantiate the data object.

        Args:
            data_path: The file path to the data Returns:
            batch_size: The batch size of the data loader tuple: shape of an example feature array
            **kwargs: Additional arguments, passed to super init and load_mnist_shard

        Returns:
           none
        """
        self.shard_num = data_path
        self.data_path = dlu.download_data_()

        self.batch_size = batch_size

        train, valid, details = dlu.load_shard(collaborator_count, self.shard_num,
                                               self.data_path, num_samples, split_ratio)

        self.num_samples = details['num_samples']
        self.num_encoder_tokens = details['num_encoder_tokens']
        self.num_decoder_tokens = details['num_decoder_tokens']
        self.max_encoder_seq_length = details['max_encoder_seq_length']
        self.max_decoder_seq_length = details['max_decoder_seq_length']

        self.X_train = [train[0], train[1]]
        self.y_train = train[2]
        self.X_valid = [valid[0], valid[1]]
        self.y_valid = valid[2]

    def get_feature_shape(self) -> Tuple[int, ...]:
        """Get the shape of an example feature array."""
        return self.X_train[0].shape

    def get_train_loader(self, batch_size: Optional[int] = None) -> Iterator[List[np.ndarray]]:
        """
        Get training data loader.

        Returns
        -------
        loader object
        """
        return self._get_batch_generator(X1=self.X_train[0], X2=self.X_train[1],
                                         y=self.y_train, batch_size=batch_size)

    def get_valid_loader(self, batch_size: Optional[int] = None) -> Iterator[List[np.ndarray]]:
        """
        Get validation data loader.

        Returns:
            loader object
        """
        return self._get_batch_generator(X1=self.X_valid[0], X2=self.X_valid[1],
                                         y=self.y_valid, batch_size=batch_size)

    def get_train_data_size(self) -> int:
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        return self.X_train[0].shape[0]

    def get_valid_data_size(self) -> int:
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        return self.X_valid[0].shape[0]

    @staticmethod
    def _batch_generator(X1: np.ndarray, X2: np.ndarray,
                         y: np.ndarray, idxs: np.ndarray,
                         batch_size: int,
                         num_batches: int) -> Iterator[List[np.ndarray]]:
        """
        Generate batch of data.

        Args:
            X: input data
            y: label data
            idxs: The index of the dataset
            batch_size: The batch size for the data loader
            num_batches: The number of batches
        Yields:
            tuple: input data, label data
        """
        for i in range(num_batches):
            a = i * batch_size
            b = a + batch_size
            yield [X1[idxs[a:b]], X2[idxs[a:b]]], y[idxs[a:b]]

    def _get_batch_generator(self, X1: np.ndarray, X2: np.ndarray,
                             y: np.ndarray,
                             batch_size: Union[int, None]) -> Iterator[List[np.ndarray]]:
        """
        Return the dataset generator.

        Args:
            X1: input data  (encoder)
            X2: input data  (decoder)
            y: label data
            batch_size: The batch size for the data loader
        """
        if batch_size is None:
            batch_size = self.batch_size
        # shuffle data indices
        idxs = np.random.permutation(np.arange(X1.shape[0]))
        # compute the number of batches
        num_batches = int(np.ceil(X1.shape[0] / batch_size))
        # build the generator and return it
        # TODO: due to _batch_generator(X1, ...) has first param X1, all params here will be moved,
        #       X1 -> X2, X2 -> y, y -> idxs, idxs -> batch_size, batch_size -> num_batches,
        #       and num_batches -> should be unexpected in this function
        return self._batch_generator(X1, X2, y, idxs, batch_size, num_batches)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Numpy optimizers package."""
from .adagrad_optimizer import NumPyAdagrad
from .adam_optimizer import NumPyAdam
from .yogi_optimizer import NumPyYogi

__all__ = [
    'NumPyAdagrad',
    'NumPyAdam',
    'NumPyYogi',
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/director.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fopenfl/protocols/director.proto\x12\x0fopenfl.director\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1bopenfl/protocols/base.proto\"\xab\x01\n\x0e\x43udaDeviceInfo\x12\r\n\x05index\x18\x01 \x01(\x04\x12\x14\n\x0cmemory_total\x18\x02 \x01(\x04\x12\x17\n\x0fmemory_utilized\x18\x03 \x01(\x04\x12\x1a\n\x12\x64\x65vice_utilization\x18\x04 \x01(\t\x12\x1b\n\x13\x63uda_driver_version\x18\x05 \x01(\t\x12\x14\n\x0c\x63uda_version\x18\x06 \x01(\t\x12\x0c\n\x04name\x18\x07 \x01(\t\"O\n\x08NodeInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x35\n\x0c\x63uda_devices\x18\x02 \x03(\x0b\x32\x1f.openfl.director.CudaDeviceInfo\"\x93\x01\n\tShardInfo\x12,\n\tnode_info\x18\x01 \x01(\x0b\x32\x19.openfl.director.NodeInfo\x12\x19\n\x11shard_description\x18\x02 \x01(\t\x12\x11\n\tn_samples\x18\x03 \x01(\x04\x12\x14\n\x0csample_shape\x18\x04 \x03(\t\x12\x14\n\x0ctarget_shape\x18\x05 \x03(\t\"H\n\x16UpdateShardInfoRequest\x12.\n\nshard_info\x18\x01 \x01(\x0b\x32\x1a.openfl.director.ShardInfo\"+\n\x17UpdateShardInfoResponse\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"2\n\x15WaitExperimentRequest\x12\x19\n\x11\x63ollaborator_name\x18\x01 \x01(\t\"1\n\x16WaitExperimentResponse\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"N\n\x18GetExperimentDataRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\x12\x19\n\x11\x63ollaborator_name\x18\x02 \x01(\t\"/\n\x0e\x45xperimentData\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x0f\n\x07npbytes\x18\x02 \x01(\x0c\"~\n\x18UpdateEnvoyStatusRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x15is_experiment_running\x18\x02 \x01(\x08\x12\x35\n\x0c\x63uda_devices\x18\x03 \x03(\x0b\x32\x1f.openfl.director.CudaDeviceInfo\"S\n\x19UpdateEnvoyStatusResponse\x12\x36\n\x13health_check_period\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x7f\n\x1aSetExperimentFailedRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\x12\x19\n\x11\x63ollaborator_name\x18\x02 \x01(\t\x12\x12\n\nerror_code\x18\x03 \x01(\r\x12\x19\n\x11\x65rror_description\x18\x04 \x01(\t\"\x1d\n\x1bSetExperimentFailedResponse\"\x1b\n\x19GetExperimentsListRequest\"x\n\x12\x45xperimentListItem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x1c\n\x14\x63ollaborators_amount\x18\x03 \x01(\r\x12\x14\n\x0ctasks_amount\x18\x04 \x01(\r\x12\x10\n\x08progress\x18\x05 \x01(\x02\"V\n\x1aGetExperimentsListResponse\x12\x38\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32#.openfl.director.ExperimentListItem\"/\n\x1fGetExperimentDescriptionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"N\n GetExperimentDescriptionResponse\x12*\n\nexperiment\x18\x01 \x01(\x0b\x32\x16.ExperimentDescription\"\x96\x01\n\x0e\x45xperimentInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x12\x63ollaborator_names\x18\x02 \x03(\t\x12\x38\n\x0f\x65xperiment_data\x18\x03 \x01(\x0b\x32\x1f.openfl.director.ExperimentData\x12 \n\x0bmodel_proto\x18\x04 \x01(\x0b\x32\x0b.ModelProto\",\n\x18SetNewExperimentResponse\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"\xa5\x01\n\x16GetTrainedModelRequest\x12\x17\n\x0f\x65xperiment_name\x18\x02 \x01(\t\x12\x45\n\nmodel_type\x18\x03 \x01(\x0e\x32\x31.openfl.director.GetTrainedModelRequest.ModelType\"+\n\tModelType\x12\x0e\n\nBEST_MODEL\x10\x00\x12\x0e\n\nLAST_MODEL\x10\x01\"8\n\x14TrainedModelResponse\x12 \n\x0bmodel_proto\x18\x01 \x01(\x0b\x32\x0b.ModelProto\"\x17\n\x15GetDatasetInfoRequest\"H\n\x16GetDatasetInfoResponse\x12.\n\nshard_info\x18\x01 \x01(\x0b\x32\x1a.openfl.director.ShardInfo\"1\n\x16GetMetricStreamRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"}\n\x17GetMetricStreamResponse\x12\x15\n\rmetric_origin\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x13\n\x0bmetric_name\x18\x03 \x01(\t\x12\x14\n\x0cmetric_value\x18\x04 \x01(\x02\x12\r\n\x05round\x18\x05 \x01(\r\"2\n\x17RemoveExperimentRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"3\n\x18RemoveExperimentResponse\x12\x17\n\x0f\x61\x63knowledgement\x18\x01 \x01(\x08\"\xeb\x01\n\tEnvoyInfo\x12.\n\nshard_info\x18\x01 \x01(\x0b\x32\x1a.openfl.director.ShardInfo\x12\x11\n\tis_online\x18\x02 \x01(\x08\x12\x1d\n\x15is_experiment_running\x18\x03 \x01(\x08\x12\x30\n\x0clast_updated\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\x0evalid_duration\x18\x05 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x17\n\x0f\x65xperiment_name\x18\x06 \x01(\t\"\x12\n\x10GetEnvoysRequest\"D\n\x11GetEnvoysResponse\x12/\n\x0b\x65nvoy_infos\x18\x01 \x03(\x0b\x32\x1a.openfl.director.EnvoyInfo2\xf4\n\n\x08\x44irector\x12\x66\n\x0fUpdateShardInfo\x12\'.openfl.director.UpdateShardInfoRequest\x1a(.openfl.director.UpdateShardInfoResponse\"\x00\x12g\n\x0eWaitExperiment\x12&.openfl.director.WaitExperimentRequest\x1a\'.openfl.director.WaitExperimentResponse\"\x00(\x01\x30\x01\x12\x63\n\x11GetExperimentData\x12).openfl.director.GetExperimentDataRequest\x1a\x1f.openfl.director.ExperimentData\"\x00\x30\x01\x12l\n\x11UpdateEnvoyStatus\x12).openfl.director.UpdateEnvoyStatusRequest\x1a*.openfl.director.UpdateEnvoyStatusResponse\"\x00\x12r\n\x13SetExperimentFailed\x12+.openfl.director.SetExperimentFailedRequest\x1a,.openfl.director.SetExperimentFailedResponse\"\x00\x12\x81\x01\n\x18GetExperimentDescription\x12\x30.openfl.director.GetExperimentDescriptionRequest\x1a\x31.openfl.director.GetExperimentDescriptionResponse\"\x00\x12o\n\x12GetExperimentsList\x12*.openfl.director.GetExperimentsListRequest\x1a+.openfl.director.GetExperimentsListResponse\"\x00\x12\x62\n\x10SetNewExperiment\x12\x1f.openfl.director.ExperimentInfo\x1a).openfl.director.SetNewExperimentResponse\"\x00(\x01\x12\x63\n\x0eGetDatasetInfo\x12&.openfl.director.GetDatasetInfoRequest\x1a\'.openfl.director.GetDatasetInfoResponse\"\x00\x12\x63\n\x0fGetTrainedModel\x12\'.openfl.director.GetTrainedModelRequest\x1a%.openfl.director.TrainedModelResponse\"\x00\x12h\n\x0fGetMetricStream\x12\'.openfl.director.GetMetricStreamRequest\x1a(.openfl.director.GetMetricStreamResponse\"\x00\x30\x01\x12m\n\x14RemoveExperimentData\x12(.openfl.director.RemoveExperimentRequest\x1a).openfl.director.RemoveExperimentResponse\"\x00\x12T\n\tGetEnvoys\x12!.openfl.director.GetEnvoysRequest\x1a\".openfl.director.GetEnvoysResponse\"\x00\x62\x06proto3')



_CUDADEVICEINFO = DESCRIPTOR.message_types_by_name['CudaDeviceInfo']
_NODEINFO = DESCRIPTOR.message_types_by_name['NodeInfo']
_SHARDINFO = DESCRIPTOR.message_types_by_name['ShardInfo']
_UPDATESHARDINFOREQUEST = DESCRIPTOR.message_types_by_name['UpdateShardInfoRequest']
_UPDATESHARDINFORESPONSE = DESCRIPTOR.message_types_by_name['UpdateShardInfoResponse']
_WAITEXPERIMENTREQUEST = DESCRIPTOR.message_types_by_name['WaitExperimentRequest']
_WAITEXPERIMENTRESPONSE = DESCRIPTOR.message_types_by_name['WaitExperimentResponse']
_GETEXPERIMENTDATAREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentDataRequest']
_EXPERIMENTDATA = DESCRIPTOR.message_types_by_name['ExperimentData']
_UPDATEENVOYSTATUSREQUEST = DESCRIPTOR.message_types_by_name['UpdateEnvoyStatusRequest']
_UPDATEENVOYSTATUSRESPONSE = DESCRIPTOR.message_types_by_name['UpdateEnvoyStatusResponse']
_SETEXPERIMENTFAILEDREQUEST = DESCRIPTOR.message_types_by_name['SetExperimentFailedRequest']
_SETEXPERIMENTFAILEDRESPONSE = DESCRIPTOR.message_types_by_name['SetExperimentFailedResponse']
_GETEXPERIMENTSLISTREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentsListRequest']
_EXPERIMENTLISTITEM = DESCRIPTOR.message_types_by_name['ExperimentListItem']
_GETEXPERIMENTSLISTRESPONSE = DESCRIPTOR.message_types_by_name['GetExperimentsListResponse']
_GETEXPERIMENTDESCRIPTIONREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionRequest']
_GETEXPERIMENTDESCRIPTIONRESPONSE = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionResponse']
_EXPERIMENTINFO = DESCRIPTOR.message_types_by_name['ExperimentInfo']
_SETNEWEXPERIMENTRESPONSE = DESCRIPTOR.message_types_by_name['SetNewExperimentResponse']
_GETTRAINEDMODELREQUEST = DESCRIPTOR.message_types_by_name['GetTrainedModelRequest']
_TRAINEDMODELRESPONSE = DESCRIPTOR.message_types_by_name['TrainedModelResponse']
_GETDATASETINFOREQUEST = DESCRIPTOR.message_types_by_name['GetDatasetInfoRequest']
_GETDATASETINFORESPONSE = DESCRIPTOR.message_types_by_name['GetDatasetInfoResponse']
_GETMETRICSTREAMREQUEST = DESCRIPTOR.message_types_by_name['GetMetricStreamRequest']
_GETMETRICSTREAMRESPONSE = DESCRIPTOR.message_types_by_name['GetMetricStreamResponse']
_REMOVEEXPERIMENTREQUEST = DESCRIPTOR.message_types_by_name['RemoveExperimentRequest']
_REMOVEEXPERIMENTRESPONSE = DESCRIPTOR.message_types_by_name['RemoveExperimentResponse']
_ENVOYINFO = DESCRIPTOR.message_types_by_name['EnvoyInfo']
_GETENVOYSREQUEST = DESCRIPTOR.message_types_by_name['GetEnvoysRequest']
_GETENVOYSRESPONSE = DESCRIPTOR.message_types_by_name['GetEnvoysResponse']
_GETTRAINEDMODELREQUEST_MODELTYPE = _GETTRAINEDMODELREQUEST.enum_types_by_name['ModelType']
CudaDeviceInfo = _reflection.GeneratedProtocolMessageType('CudaDeviceInfo', (_message.Message,), {
  'DESCRIPTOR' : _CUDADEVICEINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.CudaDeviceInfo)
  })
_sym_db.RegisterMessage(CudaDeviceInfo)

NodeInfo = _reflection.GeneratedProtocolMessageType('NodeInfo', (_message.Message,), {
  'DESCRIPTOR' : _NODEINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.NodeInfo)
  })
_sym_db.RegisterMessage(NodeInfo)

ShardInfo = _reflection.GeneratedProtocolMessageType('ShardInfo', (_message.Message,), {
  'DESCRIPTOR' : _SHARDINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ShardInfo)
  })
_sym_db.RegisterMessage(ShardInfo)

UpdateShardInfoRequest = _reflection.GeneratedProtocolMessageType('UpdateShardInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESHARDINFOREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateShardInfoRequest)
  })
_sym_db.RegisterMessage(UpdateShardInfoRequest)

UpdateShardInfoResponse = _reflection.GeneratedProtocolMessageType('UpdateShardInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESHARDINFORESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateShardInfoResponse)
  })
_sym_db.RegisterMessage(UpdateShardInfoResponse)

WaitExperimentRequest = _reflection.GeneratedProtocolMessageType('WaitExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _WAITEXPERIMENTREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.WaitExperimentRequest)
  })
_sym_db.RegisterMessage(WaitExperimentRequest)

WaitExperimentResponse = _reflection.GeneratedProtocolMessageType('WaitExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _WAITEXPERIMENTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.WaitExperimentResponse)
  })
_sym_db.RegisterMessage(WaitExperimentResponse)

GetExperimentDataRequest = _reflection.GeneratedProtocolMessageType('GetExperimentDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDATAREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentDataRequest)
  })
_sym_db.RegisterMessage(GetExperimentDataRequest)

ExperimentData = _reflection.GeneratedProtocolMessageType('ExperimentData', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTDATA,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ExperimentData)
  })
_sym_db.RegisterMessage(ExperimentData)

UpdateEnvoyStatusRequest = _reflection.GeneratedProtocolMessageType('UpdateEnvoyStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENVOYSTATUSREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateEnvoyStatusRequest)
  })
_sym_db.RegisterMessage(UpdateEnvoyStatusRequest)

UpdateEnvoyStatusResponse = _reflection.GeneratedProtocolMessageType('UpdateEnvoyStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENVOYSTATUSRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateEnvoyStatusResponse)
  })
_sym_db.RegisterMessage(UpdateEnvoyStatusResponse)

SetExperimentFailedRequest = _reflection.GeneratedProtocolMessageType('SetExperimentFailedRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETEXPERIMENTFAILEDREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.SetExperimentFailedRequest)
  })
_sym_db.RegisterMessage(SetExperimentFailedRequest)

SetExperimentFailedResponse = _reflection.GeneratedProtocolMessageType('SetExperimentFailedResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETEXPERIMENTFAILEDRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.SetExperimentFailedResponse)
  })
_sym_db.RegisterMessage(SetExperimentFailedResponse)

GetExperimentsListRequest = _reflection.GeneratedProtocolMessageType('GetExperimentsListRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTSLISTREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentsListRequest)
  })
_sym_db.RegisterMessage(GetExperimentsListRequest)

ExperimentListItem = _reflection.GeneratedProtocolMessageType('ExperimentListItem', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTLISTITEM,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ExperimentListItem)
  })
_sym_db.RegisterMessage(ExperimentListItem)

GetExperimentsListResponse = _reflection.GeneratedProtocolMessageType('GetExperimentsListResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTSLISTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentsListResponse)
  })
_sym_db.RegisterMessage(GetExperimentsListResponse)

GetExperimentDescriptionRequest = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentDescriptionRequest)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionRequest)

GetExperimentDescriptionResponse = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentDescriptionResponse)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionResponse)

ExperimentInfo = _reflection.GeneratedProtocolMessageType('ExperimentInfo', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ExperimentInfo)
  })
_sym_db.RegisterMessage(ExperimentInfo)

SetNewExperimentResponse = _reflection.GeneratedProtocolMessageType('SetNewExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETNEWEXPERIMENTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.SetNewExperimentResponse)
  })
_sym_db.RegisterMessage(SetNewExperimentResponse)

GetTrainedModelRequest = _reflection.GeneratedProtocolMessageType('GetTrainedModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAINEDMODELREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetTrainedModelRequest)
  })
_sym_db.RegisterMessage(GetTrainedModelRequest)

TrainedModelResponse = _reflection.GeneratedProtocolMessageType('TrainedModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _TRAINEDMODELRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.TrainedModelResponse)
  })
_sym_db.RegisterMessage(TrainedModelResponse)

GetDatasetInfoRequest = _reflection.GeneratedProtocolMessageType('GetDatasetInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETINFOREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetDatasetInfoRequest)
  })
_sym_db.RegisterMessage(GetDatasetInfoRequest)

GetDatasetInfoResponse = _reflection.GeneratedProtocolMessageType('GetDatasetInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETINFORESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetDatasetInfoResponse)
  })
_sym_db.RegisterMessage(GetDatasetInfoResponse)

GetMetricStreamRequest = _reflection.GeneratedProtocolMessageType('GetMetricStreamRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetMetricStreamRequest)
  })
_sym_db.RegisterMessage(GetMetricStreamRequest)

GetMetricStreamResponse = _reflection.GeneratedProtocolMessageType('GetMetricStreamResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetMetricStreamResponse)
  })
_sym_db.RegisterMessage(GetMetricStreamResponse)

RemoveExperimentRequest = _reflection.GeneratedProtocolMessageType('RemoveExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEEXPERIMENTREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.RemoveExperimentRequest)
  })
_sym_db.RegisterMessage(RemoveExperimentRequest)

RemoveExperimentResponse = _reflection.GeneratedProtocolMessageType('RemoveExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEEXPERIMENTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.RemoveExperimentResponse)
  })
_sym_db.RegisterMessage(RemoveExperimentResponse)

EnvoyInfo = _reflection.GeneratedProtocolMessageType('EnvoyInfo', (_message.Message,), {
  'DESCRIPTOR' : _ENVOYINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.EnvoyInfo)
  })
_sym_db.RegisterMessage(EnvoyInfo)

GetEnvoysRequest = _reflection.GeneratedProtocolMessageType('GetEnvoysRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETENVOYSREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetEnvoysRequest)
  })
_sym_db.RegisterMessage(GetEnvoysRequest)

GetEnvoysResponse = _reflection.GeneratedProtocolMessageType('GetEnvoysResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETENVOYSRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetEnvoysResponse)
  })
_sym_db.RegisterMessage(GetEnvoysResponse)

_DIRECTOR = DESCRIPTOR.services_by_name['Director']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CUDADEVICEINFO._serialized_start=147
  _CUDADEVICEINFO._serialized_end=318
  _NODEINFO._serialized_start=320
  _NODEINFO._serialized_end=399
  _SHARDINFO._serialized_start=402
  _SHARDINFO._serialized_end=549
  _UPDATESHARDINFOREQUEST._serialized_start=551
  _UPDATESHARDINFOREQUEST._serialized_end=623
  _UPDATESHARDINFORESPONSE._serialized_start=625
  _UPDATESHARDINFORESPONSE._serialized_end=668
  _WAITEXPERIMENTREQUEST._serialized_start=670
  _WAITEXPERIMENTREQUEST._serialized_end=720
  _WAITEXPERIMENTRESPONSE._serialized_start=722
  _WAITEXPERIMENTRESPONSE._serialized_end=771
  _GETEXPERIMENTDATAREQUEST._serialized_start=773
  _GETEXPERIMENTDATAREQUEST._serialized_end=851
  _EXPERIMENTDATA._serialized_start=853
  _EXPERIMENTDATA._serialized_end=900
  _UPDATEENVOYSTATUSREQUEST._serialized_start=902
  _UPDATEENVOYSTATUSREQUEST._serialized_end=1028
  _UPDATEENVOYSTATUSRESPONSE._serialized_start=1030
  _UPDATEENVOYSTATUSRESPONSE._serialized_end=1113
  _SETEXPERIMENTFAILEDREQUEST._serialized_start=1115
  _SETEXPERIMENTFAILEDREQUEST._serialized_end=1242
  _SETEXPERIMENTFAILEDRESPONSE._serialized_start=1244
  _SETEXPERIMENTFAILEDRESPONSE._serialized_end=1273
  _GETEXPERIMENTSLISTREQUEST._serialized_start=1275
  _GETEXPERIMENTSLISTREQUEST._serialized_end=1302
  _EXPERIMENTLISTITEM._serialized_start=1304
  _EXPERIMENTLISTITEM._serialized_end=1424
  _GETEXPERIMENTSLISTRESPONSE._serialized_start=1426
  _GETEXPERIMENTSLISTRESPONSE._serialized_end=1512
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_start=1514
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_end=1561
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_start=1563
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_end=1641
  _EXPERIMENTINFO._serialized_start=1644
  _EXPERIMENTINFO._serialized_end=1794
  _SETNEWEXPERIMENTRESPONSE._serialized_start=1796
  _SETNEWEXPERIMENTRESPONSE._serialized_end=1840
  _GETTRAINEDMODELREQUEST._serialized_start=1843
  _GETTRAINEDMODELREQUEST._serialized_end=2008
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_start=1965
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_end=2008
  _TRAINEDMODELRESPONSE._serialized_start=2010
  _TRAINEDMODELRESPONSE._serialized_end=2066
  _GETDATASETINFOREQUEST._serialized_start=2068
  _GETDATASETINFOREQUEST._serialized_end=2091
  _GETDATASETINFORESPONSE._serialized_start=2093
  _GETDATASETINFORESPONSE._serialized_end=2165
  _GETMETRICSTREAMREQUEST._serialized_start=2167
  _GETMETRICSTREAMREQUEST._serialized_end=2216
  _GETMETRICSTREAMRESPONSE._serialized_start=2218
  _GETMETRICSTREAMRESPONSE._serialized_end=2343
  _REMOVEEXPERIMENTREQUEST._serialized_start=2345
  _REMOVEEXPERIMENTREQUEST._serialized_end=2395
  _REMOVEEXPERIMENTRESPONSE._serialized_start=2397
  _REMOVEEXPERIMENTRESPONSE._serialized_end=2448
  _ENVOYINFO._serialized_start=2451
  _ENVOYINFO._serialized_end=2686
  _GETENVOYSREQUEST._serialized_start=2688
  _GETENVOYSREQUEST._serialized_end=2706
  _GETENVOYSRESPONSE._serialized_start=2708
  _GETENVOYSRESPONSE._serialized_end=2776
  _DIRECTOR._serialized_start=2779
  _DIRECTOR._serialized_end=4175
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Dill serializer plugin."""

import dill

from .serializer_interface import Serializer


class DillSerializer(Serializer):
    """Serializer API plugin."""

    def __init__(self) -> None:
        """Initialize serializer."""
        super().__init__()

    @staticmethod
    def serialize(object_, filename):
        """Serialize an object and save to disk."""
        with open(filename, 'wb') as f:
            dill.dump(object_, f, recurse=True)

    @staticmethod
    def restore_object(filename):
        """Load and deserialize an object."""
        with open(filename, 'rb') as f:
            return dill.load(f)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""DataLoader module."""


class DataLoader:
    """Federated Learning Data Loader Class."""

    def __init__(self, **kwargs):
        """
        Instantiate the data object.

        Returns:
            None
        """
        pass

    def get_feature_shape(self):
        """
        Get the shape of an example feature array.

        Returns:
            tuple: shape of an example feature array
        """
        raise NotImplementedError

    def get_train_loader(self, **kwargs):
        """
        Get training data loader.

        Returns:
            loader object (class defined by inheritor)
        """
        raise NotImplementedError

    def get_valid_loader(self):
        """
        Get validation data loader.

        Returns:
            loader object (class defined by inheritor)
        """
        raise NotImplementedError

    def get_infer_loader(self):
        """
        Get inferencing data loader.

        Returns
        -------
        loader object (class defined by inheritor)
        """
        return NotImplementedError

    def get_train_data_size(self):
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        raise NotImplementedError

    def get_valid_data_size(self):
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        raise NotImplementedError



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""FastEstimatorTaskRunner module."""

import numpy as np

from openfl.utilities import change_tags
from openfl.utilities import split_tensor_dict_for_holdouts
from openfl.utilities import TensorKey
from .runner import TaskRunner
from .runner_keras import KerasTaskRunner
from .runner_pt import PyTorchTaskRunner


class FastEstimatorTaskRunner(TaskRunner):
    """A wrapper for fastestimator.estimator."""

    def __init__(self, estimator, **kwargs):
        """Initialize.

        Args:
            estimator: object of type fastestimator.estimator
        """
        super().__init__(**kwargs)
        import fastestimator as fe

        class ProgressLoader(fe.trace.Trace):
            def __init__(self, get_progress) -> None:
                super().__init__(mode='train')
                self.get_progress = get_progress

            def on_begin(self, data) -> None:
                """Run once at the beginning of training or testing.

                Args:
                    data: A dictionary through which traces can communicate with
                    each other or write values for logging.
                """
                progress = self.get_progress()
                self.system.epoch_idx = progress['epoch_idx']
                self.system.global_step = progress['global_step']

        estimator_kwargs = {}
        for k, v in estimator.system.__dict__.items():
            if k in ['pipeline', 'network', 'log_steps',
                     'max_train_steps_per_epoch', 'max_eval_steps_per_epoch']:
                estimator_kwargs[k] = v
            if k == 'traces':
                self.logger.debug(f'traces={estimator.system.traces}')
                estimator_kwargs[k] = v + [
                    ProgressLoader(lambda: {
                        'epoch_idx': self.epoch_idx,
                        'global_step': self.global_step
                    })]
        estimator_kwargs.update({
            'epochs': estimator.system.total_epochs,
            'monitor_names': estimator.monitor_names
        })
        self.estimator = fe.Estimator(**estimator_kwargs)
        assert (len(estimator.network.models) == 1), (
            'Only one-model networks are currently supported')
        if isinstance(estimator.network, fe.network.TorchNetwork):
            impl = PyTorchTaskRunner
        elif isinstance(estimator.network, fe.network.TFNetwork):
            impl = KerasTaskRunner
        self.model = self.estimator.network.models[0]
        self.optimizer = self.model.optimizer
        self.runner = impl(**kwargs)
        self.runner.model = self.model
        self.runner.optimizer = self.optimizer
        self.required_tensorkeys_for_function = {}
        self.tensor_dict_split_fn_kwargs = self.runner.tensor_dict_split_fn_kwargs
        self.initialize_tensorkeys_for_functions()
        self.epoch_idx = 0
        self.global_step = None
        self.total_epochs = self.estimator.system.total_epochs

    def train(self, col_name, round_num, input_tensor_dict, epochs, **kwargs):
        """Perform training for a specified number of epochs."""
        if 'metrics' not in kwargs:
            raise KeyError('metrics must be included in kwargs')
        param_metrics = kwargs['metrics']

        self.rebuild_model(round_num, input_tensor_dict)

        # Estimators need to be given an experiment name to produce an output
        # summary
        summary = self.estimator.fit('experiment', warmup=False)
        self.epoch_idx = self.estimator.system.epoch_idx
        self.global_step = self.estimator.system.global_step
        self.estimator.system.total_epochs += self.total_epochs
        # Define what the ouptut is to encapsulate in tensorkeys and return
        # output metric tensors (scalar)
        origin = col_name
        tags = ('trained',)
        output_metric_dict = {
            TensorKey(
                metric, origin, round_num, True, ('metric',)
            ): np.array(
                list(
                    summary.history['train'][metric].values()
                )[-1]) for metric in param_metrics}

        # output model tensors (Doesn't include TensorKey)
        output_model_dict = self.get_tensor_dict(with_opt_vars=True)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            self.logger, output_model_dict,
            **self.tensor_dict_split_fn_kwargs
        )

        # create global tensorkeys
        global_tensorkey_model_dict = {
            TensorKey(
                tensor_name, origin, round_num, False, tags
            ): nparray for tensor_name, nparray in global_model_dict.items()
        }
        # create tensorkeys that should stay local
        local_tensorkey_model_dict = {
            TensorKey(
                tensor_name, origin, round_num, False, tags
            ): nparray for tensor_name, nparray in local_model_dict.items()
        }
        # the train/validate aggregated function of the next round will look
        # for the updated model parameters.
        # this ensures they will be resolved locally
        next_local_tensorkey_model_dict = {
            TensorKey(
                tensor_name, origin, round_num + 1, False, ('model',)
            ): nparray for tensor_name, nparray in local_model_dict.items()
        }

        global_tensor_dict = {
            **output_metric_dict,
            **global_tensorkey_model_dict
        }
        local_tensor_dict = {
            **local_tensorkey_model_dict,
            **next_local_tensorkey_model_dict
        }

        # update the required tensors if they need to be pulled from the
        # aggregator
        # TODO this logic can break if different collaborators have different
        #  roles between rounds.
        # for example, if a collaborator only performs validation in the first
        # round but training in the second, it has no way of knowing the
        # optimizer state tensor names to request from the aggregator
        # because these are only created after training occurs.
        # A work around could involve doing a single epoch of training
        # on random data to get the optimizer names, and then throwing away
        # the model.
        if self.opt_treatment == 'CONTINUE_GLOBAL':
            self.initialize_tensorkeys_for_functions(with_opt_vars=True)

        return global_tensor_dict, local_tensor_dict

    def validate(self, col_name, round_num, input_tensor_dict, **kwargs):
        """
        Run the trained model on validation data; report results.

        Parameters
        ----------
        input_tensor_dict : either the last aggregated or locally trained model

        Returns
        -------
        output_tensor_dict : {TensorKey: nparray} (these correspond to acc,
         precision, f1_score, etc.)
        """
        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        param_metrics = kwargs['metrics']

        results = self.estimator.test('experiment')
        ret_dict = {
            metric: list(results.history['test'][metric].values())[-1]
            for metric in param_metrics
        }

        origin = col_name
        suffix = 'validate'
        if kwargs['apply'] == 'local':
            suffix += '_local'
        else:
            suffix += '_agg'
        tags = ('metric',)
        tags = change_tags(tags, add_field=suffix)
        output_tensor_dict = {
            TensorKey(
                metric, origin, round_num, True, tags
            ): np.array(ret_dict[metric])
            for metric in param_metrics
        }

        return output_tensor_dict, {}

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """
        Set the required tensors for all publicly accessible methods that could \
            be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.
         Custom tensors should be added to this function

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.runner.initialize_tensorkeys_for_functions(with_opt_vars)

    def build_model(self):
        """Abstract method."""
        raise NotImplementedError

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """
        When running a task, a map of named tensorkeys must be provided to the \
            function as dependencies.

        Returns:
            list: (TensorKey(tensor_name, origin, round_number))
        """
        return self.runner.get_required_tensorkeys_for_function(
            func_name, **kwargs)

    def get_tensor_dict(self, with_opt_vars):
        """
        Get the weights.

        Args:
            with_opt_vars (bool): Specify if we also want to get the variables
             of the optimizer.

        Returns:
            dict: The weight dictionary {<tensor_name>: <value>}
        """
        return self.runner.get_tensor_dict(with_opt_vars)

    def set_tensor_dict(self, tensor_dict, with_opt_vars):
        """
        Set the model weights with a tensor dictionary: {<tensor_name>: <value>}.

        Args:
            tensor_dict (dict): The model weights dictionary.
            with_opt_vars (bool): Specify if we also want to set the variables
            of the optimizer.

        Returns:
            None
        """
        return self.runner.set_tensor_dict(tensor_dict, with_opt_vars)

    def reset_opt_vars(self):
        """Reinitialize the optimizer variables."""
        return self.runner.reset_opt_vars()

    def initialize_globals(self):
        """
        Initialize all global variables.

        Returns:
            None
        """
        return self.runner.initialize_globals()

    def load_native(self, filepath, **kwargs):
        """
        Load model state from a filepath in ML-framework "native" format, \
            e.g. PyTorch pickled models.

        May load from multiple files. Other filepaths may be derived from the
        passed filepath, or they may be in the kwargs.

        Args:
            filepath (string): Path to frame-work specific file to load. For
                               frameworks that use multiple files, this string
                               must be used to derive the other filepaths.
            kwargs           : For future-proofing

        Returns:
            None
        """
        return self.runner.load_native(filepath, **kwargs)

    def save_native(self, filepath, **kwargs):
        """
        Save model state in ML-framework "native" format, \
            e.g. PyTorch pickled models.

        May save one file or multiple files, depending on the framework.

        Args:
            filepath (string): If framework stores a single file, this should
                               be a single file path.
            Frameworks that store multiple files may need to derive the other
            paths from this path.
            kwargs           : For future-proofing

        Returns:
            None
        """
        return self.runner.save_native(filepath, **kwargs)

    def rebuild_model(self, round_num, input_tensor_dict, validation=False):
        """
        Parse tensor names and update weights of model. Handles the optimizer treatment.

        Returns:
            None
        """
        return self.runner.rebuild_model(
            round_num, input_tensor_dict, validation)

    def set_optimizer_treatment(self, opt_treatment):
        """Change treatment of current instance optimizer."""
        super().set_optimizer_treatment(opt_treatment)
        self.runner.opt_treatment = opt_treatment



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Tutorial module."""

from logging import getLogger

from click import group
from click import IntRange
from click import option
from click import pass_context

from openfl.utilities import click_types

logger = getLogger(__name__)


@group()
@pass_context
def tutorial(context):
    """Manage Jupyter notebooks."""
    context.obj['group'] = 'tutorial'


@tutorial.command()
@option('-ip', '--ip', required=False, type=click_types.IP_ADDRESS,
        help='IP address the Jupyter Lab that should start')
@option('-port', '--port', required=False, type=IntRange(1, 65535),
        help='The port the Jupyter Lab server will listen on')
def start(ip, port):
    """Start the Jupyter Lab from the tutorials directory."""
    from os import environ
    from subprocess import check_call
    from sys import executable

    from openfl.interface.cli_helper import TUTORIALS

    if 'VIRTUAL_ENV' in environ:
        venv = environ['VIRTUAL_ENV'].split('/')[-1]
        check_call([
            executable, '-m', 'ipykernel', 'install',
            '--user', '--name', f'{venv}'
        ], shell=False)

    jupyter_command = ['jupyter', 'lab', '--notebook-dir', f'{TUTORIALS}']

    if ip is not None:
        jupyter_command += ['--ip', f'{ip}']
    if port is not None:
        jupyter_command += ['--port', f'{port}']

    check_call(jupyter_command)



max_metadata_size = 32 * 2 ** 20
max_message_length = 2 ** 30

channel_options = [
    ('grpc.max_metadata_size', max_metadata_size),
    ('grpc.max_send_message_length', max_message_length),
    ('grpc.max_receive_message_length', max_message_length)
]



#!/usr/bin/env python
# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""CLI module."""

from click import argument
from click import command
from click import echo
from click import Group
from click import group
from click import option
from click import pass_context
from click import style

from openfl.utilities import add_log_level


def setup_logging(level='info', log_file=None):
    """Initialize logging settings."""
    import logging
    from logging import basicConfig

    from rich.console import Console
    from rich.logging import RichHandler

    metric = 25
    add_log_level('METRIC', metric)

    if isinstance(level, str):
        level = level.upper()

    handlers = []
    if log_file:
        fh = logging.FileHandler(log_file)
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)s %(message)s %(filename)s:%(lineno)d'
        )
        fh.setFormatter(formatter)
        handlers.append(fh)

    console = Console(width=160)
    handlers.append(RichHandler(console=console))
    basicConfig(level=level, format='%(message)s',
                datefmt='[%X]', handlers=handlers)


def disable_warnings():
    """Disables CUDA warnings."""
    import os

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'


class CLI(Group):
    """CLI class."""

    def __init__(self, name=None, commands=None, **kwargs):
        """Initialize."""
        super(CLI, self).__init__(name, commands, **kwargs)
        self.commands = commands or {}

    def list_commands(self, ctx):
        """Display all available commands."""
        return self.commands

    def format_help(self, ctx, formatter):
        """Dislpay user-friendly help."""
        show_header()
        uses = [
            f'{ctx.command_path}',
            '[options]',
            style('[command]', fg='blue'),
            style('[subcommand]', fg='cyan'),
            '[args]'
        ]

        formatter.write(style('BASH COMPLETE ACTIVATION\n\n', bold=True, fg='bright_black'))
        formatter.write(
            'Run in terminal:\n'
            '   _FX_COMPLETE=bash_source fx > ~/.fx-autocomplete.sh\n'
            '   source ~/.fx-autocomplete.sh\n'
            'If ~/.fx-autocomplete.sh has already exist:\n'
            '   source ~/.fx-autocomplete.sh\n\n'
        )

        formatter.write(style('CORRECT USAGE\n\n', bold=True, fg='bright_black'))
        formatter.write(' '.join(uses) + '\n')

        opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is not None:
                opts.append(rv)

        formatter.write(style(
            '\nGLOBAL OPTIONS\n\n', bold=True, fg='bright_black'))
        formatter.write_dl(opts)

        cmds = []
        for cmd in self.list_commands(ctx):
            cmd = self.get_command(ctx, cmd)
            cmds.append((cmd.name, cmd, 0))

            for sub in cmd.list_commands(ctx):
                sub = cmd.get_command(ctx, sub)
                cmds.append((sub.name, sub, 1))

        formatter.write(style(
            '\nAVAILABLE COMMANDS\n', bold=True, fg='bright_black'))

        for name, cmd, level in cmds:
            help_str = cmd.get_short_help_str()
            if level == 0:
                formatter.write(
                    f'\n{style(name, fg="blue", bold=True):<30}'
                    f' {style(help_str, bold=True)}' + '\n')
                formatter.write('─' * 80 + '\n')
            if level == 1:
                formatter.write(
                    f'  {style("*", fg="green")}'
                    f' {style(name, fg="cyan"):<21} {help_str}' + '\n')


@group(cls=CLI)
@option('-l', '--log-level', default='info', help='Logging verbosity level.')
@pass_context
def cli(context, log_level):
    """Command-line Interface."""
    import os
    from sys import argv

    context.ensure_object(dict)
    context.obj['log_level'] = log_level
    context.obj['fail'] = False
    context.obj['script'] = argv[0]
    context.obj['arguments'] = argv[1:]

    log_file = os.getenv('LOG_FILE')
    setup_logging(log_level, log_file)


@cli.result_callback()
@pass_context
def end(context, result, **kwargs):
    """Print the result of the operation."""
    if context.obj['fail']:
        echo('\n ❌ :(')
    else:
        echo('\n ✔️ OK')


@command(name='help')
@pass_context
@argument('subcommand', required=False)
def help_(context, subcommand):
    """Display help."""
    pass


def error_handler(error):
    """Handle the error."""
    if 'cannot import' in str(error):
        if 'TensorFlow' in str(error):
            echo(style('EXCEPTION', fg='red', bold=True) + ' : ' + style(
                'Tensorflow must be installed prior to running this command',
                fg='red'))
        if 'PyTorch' in str(error):
            echo(style('EXCEPTION', fg='red', bold=True) + ' : ' + style(
                'Torch must be installed prior to running this command',
                fg='red'))
    echo(style('EXCEPTION', fg='red', bold=True)
         + ' : ' + style(f'{error}', fg='red'))
    raise error


def show_header():
    """Show header."""
    banner = 'OpenFL - Open Federated Learning'
    echo(style(f'{banner:<80}', bold=True, bg='bright_blue'))
    echo()


def entry():
    """Entry point of the Command-Line Interface."""
    from importlib import import_module
    from pathlib import Path
    from sys import path

    file = Path(__file__).resolve()
    root = file.parent.resolve()  # interface root, containing command modules
    work = Path.cwd().resolve()
    path.append(str(root))
    path.insert(0, str(work))

    # Setup logging immediately to suppress unnecessary warnings on import
    # This will be overridden later with user selected debugging level
    disable_warnings()

    for module in root.glob('*.py'):  # load command modules

        package = module.parent
        module = module.name.split('.')[0]

        if module.count('__init__') or module.count('cli'):
            continue

        command_group = import_module(module, package)

        cli.add_command(command_group.__getattribute__(module))

    try:
        cli()
    except Exception as e:
        error_handler(e)


if __name__ == '__main__':
    entry()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""RandomShiftPipeline module."""

import numpy as np

from .pipeline import Float32NumpyArrayToBytes
from .pipeline import TransformationPipeline
from .pipeline import Transformer


class RandomShiftTransformer(Transformer):
    """Random Shift Transformer."""

    def __init__(self):
        """Initialize."""
        self.lossy = False

        return

    def forward(self, data, **kwargs):
        """Forward pass - compress data.

        Implement the data transformation.

        Args:
            data: an array value from a model tensor_dict

        Returns:
            transformed_data:
            metadata:

        """
        shape = data.shape
        random_shift = np.random.uniform(
            low=-20, high=20, size=shape).astype(np.float32)
        transformed_data = data + random_shift

        # construct metadata
        metadata = {'int_to_float': {}, 'int_list': list(shape)}
        for idx, val in enumerate(random_shift.flatten(order='C')):
            metadata['int_to_float'][idx] = val

        return transformed_data, metadata

    def backward(self, data, metadata, **kwargs):
        """Backward pass - Decompress data.

        Implement the data transformation needed when going the oppposite
        direction to the forward method.

        Args:
            data:
            metadata:

        Returns:
            transformed_data:
        """
        shape = tuple(metadata['int_list'])
        # this is an awkward use of the metadata into to float dict, usually
        # it will trully be treated as a dict. Here (and in 'forward' above)
        # we use it essentially as an array.
        shift = np.reshape(
            np.array([
                metadata['int_to_float'][idx]
                for idx in range(len(metadata['int_to_float']))]),
            newshape=shape,
            order='C'
        )
        return data - shift


class RandomShiftPipeline(TransformationPipeline):
    """Random Shift Pipeline."""

    def __init__(self, **kwargs):
        """Initialize."""
        transformers = [RandomShiftTransformer(), Float32NumpyArrayToBytes()]
        super(RandomShiftPipeline, self).__init__(transformers=transformers)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from openfl.federated import PyTorchDataLoader
from .histology_utils import load_histology_shard


class PyTorchHistologyInMemory(PyTorchDataLoader):
    """PyTorch data loader for Histology dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """Instantiate the data object.

        Args:
            data_path: The file path to the data
            batch_size: The batch size of the data loader
            **kwargs: Additional arguments, passed to super init
             and load_mnist_shard
        """
        super().__init__(batch_size, random_seed=0, **kwargs)

        _, num_classes, X_train, y_train, X_valid, y_valid = load_histology_shard(
            shard_num=int(data_path), **kwargs)

        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

        self.num_classes = num_classes



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Landmarks Shard Descriptor."""

import json
import shutil
from hashlib import md5
from logging import getLogger
from pathlib import Path
from random import shuffle
from typing import Dict
from typing import List
from zipfile import ZipFile

import numpy as np
import pandas as pd
from kaggle.api.kaggle_api_extended import KaggleApi

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = getLogger(__name__)


class LandmarkShardDataset(ShardDataset):
    """Landmark Shard dataset class."""

    def __init__(self, dataset_dir: Path,
                 rank: int = 1, worldsize: int = 1) -> None:
        """Initialize LandmarkShardDataset."""
        self.rank = rank
        self.worldsize = worldsize
        self.dataset_dir = dataset_dir
        self.img_names = list(self.dataset_dir.glob('img_*.npy'))

        # Sharding
        self.img_names = self.img_names[self.rank - 1::self.worldsize]
        # Shuffling the results dataset after choose half pictures of each class
        shuffle(self.img_names)

    def __getitem__(self, index) -> np.ndarray:
        """Return a item by the index."""
        # Get name key points file
        # f.e. image name:  'img_123.npy, corresponding name of the key points: 'keypoints_123.npy'
        kp_name = str(self.img_names[index]).replace('img', 'keypoints')
        return np.load(self.img_names[index]), np.load(self.dataset_dir / kp_name)

    def __len__(self) -> int:
        """Return the len of the dataset."""
        return len(self.img_names)


class LandmarkShardDescriptor(ShardDescriptor):
    """Landmark Shard descriptor class."""

    def __init__(self, data_folder: str = 'data',
                 rank_worldsize: str = '1, 1',
                 **kwargs) -> None:
        """Initialize LandmarkShardDescriptor."""
        super().__init__()
        # Settings for sharding the dataset
        self.rank, self.worldsize = map(int, rank_worldsize.split(','))

        self.data_folder = Path.cwd() / data_folder
        self.download_data()

        # Calculating data and target shapes
        ds = self.get_dataset()
        sample, target = ds[0]
        self._sample_shape = [str(dim) for dim in sample.shape]
        self._target_shape = [str(len(target.shape))]

        if self._target_shape[0] != '1':
            raise ValueError('Target has a wrong shape')

    def process_data(self, name_csv_file) -> None:
        """Process data from csv to numpy format and save it in the same folder."""
        data_df = pd.read_csv(self.data_folder / name_csv_file)
        data_df.fillna(method='ffill', inplace=True)
        keypoints = data_df.drop('Image', axis=1)
        cur_folder = self.data_folder.relative_to(Path.cwd())

        for i in range(data_df.shape[0]):
            img = data_df['Image'][i].split(' ')
            img = np.array(['0' if x == '' else x for x in img], dtype='float32').reshape(96, 96)
            np.save(str(cur_folder / f'img_{i}.npy'), img)
            y = np.array(keypoints.iloc[i, :], dtype='float32')
            np.save(str(cur_folder / f'keypoints_{i}.npy'), y)

    def download_data(self) -> None:
        """Download dataset from Kaggle."""
        if self.is_dataset_complete():
            return

        self.data_folder.mkdir(parents=True, exist_ok=True)

        logger.info('Your dataset is absent or damaged. Downloading ... ')
        api = KaggleApi()
        api.authenticate()

        if Path('data').exists():
            shutil.rmtree('data')

        api.competition_download_file(
            'facial-keypoints-detection',
            'training.zip', path=self.data_folder
        )

        with ZipFile(self.data_folder / 'training.zip', 'r') as zipobj:
            zipobj.extractall(self.data_folder)

        (self.data_folder / 'training.zip').unlink()

        self.process_data('training.csv')
        (self.data_folder / 'training.csv').unlink()
        self.save_all_md5()

    def get_dataset(self, dataset_type='train') -> LandmarkShardDataset:
        """Return a shard dataset by type."""
        return LandmarkShardDataset(
            dataset_dir=self.data_folder,
            rank=self.rank,
            worldsize=self.worldsize
        )

    def calc_all_md5(self) -> Dict[str, str]:
        """Calculate hash of all dataset."""
        md5_dict = {}
        for root in self.data_folder.glob('*.npy'):
            md5_calc = md5(usedforsecurity=False)
            rel_file = root.relative_to(self.data_folder)

            with open(self.data_folder / rel_file, 'rb') as f:
                for chunk in iter(lambda: f.read(4096), b''):
                    md5_calc.update(chunk)
                md5_dict[str(rel_file)] = md5_calc.hexdigest()
        return md5_dict

    def save_all_md5(self) -> None:
        """Save dataset hash."""
        all_md5 = self.calc_all_md5()
        with open(self.data_folder / 'dataset.json', 'w') as f:
            json.dump(all_md5, f)

    def is_dataset_complete(self) -> bool:
        """Check dataset integrity."""
        dataset_md5_path = self.data_folder / 'dataset.json'
        if dataset_md5_path.exists():
            with open(dataset_md5_path, 'r') as f:
                old_md5 = json.load(f)
            new_md5 = self.calc_all_md5()
            return new_md5 == old_md5
        return False

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        return self._sample_shape

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        return self._target_shape

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Dogs and Cats dataset, shard number {self.rank} '
                f'out of {self.worldsize}')



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregation functions experimental package."""

from .privileged_aggregation import PrivilegedAggregationFunction

__all__ = ['PrivilegedAggregationFunction',]
